/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
*/
package spatial.tree.lab;

import java.util.ArrayList;
import java.util.List;

/**
 *
 * @author Lawal 
 */
public class MinBoundingRectangle {
    public class Builder {
        public Builder setDim(int dim) {
            
            for(int i = intervals.size(); i < dim; i++)
            {
                intervals.add(new Interval());
            }

            return this;
        }

        public Builder setInterval(int dim, double a, double b) {
            MinBoundingRectangle.this.setIntervalA(dim, a);
            MinBoundingRectangle.this.setIntervalB(dim, b);
            return this;
        }

        public Builder setInterval(int dim, double v) {
            MinBoundingRectangle.this.setIntervalA(dim, v);
            MinBoundingRectangle.this.setIntervalB(dim, v);
            return this;
        }
        
        public MinBoundingRectangle build() {
            return MinBoundingRectangle.this;
        }
    }

    List<Interval> intervals = new ArrayList<>();
    List<Interval> getIntervals() {
        return intervals;
    }
    public void setIntervals(List<Interval> intervals) {
        this.dim = intervals.size();
        this.intervals = intervals;
    }
    
    int dim;
    public int getDim() {
        return dim;
    }
    void setDim(int dim) {
        this.dim = dim;
    }
     
    public MinBoundingRectangle(int dim) {
        this.dim = dim;
                
        for(int i = 0; i < dim; i++)
        {
            intervals.add(new Interval());
        }
    }
    
    public MinBoundingRectangle parent;
    public MinBoundingRectangle getParent() {
        return parent;
    }

    public void setParent(MinBoundingRectangle parent) {
        this.parent = parent;
    }
    
    public Interval getInterval(int dim) {
        return intervals.get(dim);
    }

    void setInterval(int dim, Interval interval) {
        intervals.set(dim, interval);
    }

    public double getIntervalA(int dim) {
        return intervals.get(dim).getA();
    }
    
    void setIntervalA(int dim, double a) {
        intervals.get(dim).setA(a);
    }

    public double getIntervalB(int dim) {
        return intervals.get(dim).getB();
    }
    
    void setIntervalB(int dim, double b) {
        intervals.get(dim).setB(b);
    }

    public double getIntervalC(int dim) {
        return intervals.get(dim).getC();
    }

    void setForAllAB(double a, double b)
    {
        for (Interval interval : intervals) {
            interval.setAB(a, b);
        }
    }

    double getArea()
    {
        double result;
        
        result = 1.0;
        for(Interval interval :intervals)
        {
            result = result * (interval.getB() - interval.getA());
        }
        
        return result;
    }
    
    double getMargin()
    {
        double result;
        
        result = 0.0;
        for(Interval interval :intervals)
        {
            result = result + (interval.getB() - interval.getA());
        }
        
        return result;
    }

    boolean isOverlapped(MinBoundingRectangle minBoundingRectangle)
    {
        boolean overlapped = true;
        
        for(int i=0; i < dim; i++)
        {    
            if(minBoundingRectangle.getIntervalA(i) > getIntervalB(i)) {
                overlapped = false;
            }

            if(minBoundingRectangle.getIntervalB(i) < getIntervalA(i)) {
                overlapped = false;
            }
        }
    
        return overlapped;
    }
    
    double getOverlapArea(MinBoundingRectangle minBoundingRectangle)
    {
        if(isOverlapped(minBoundingRectangle)) {
            MinBoundingRectangle overlapBoundingRectangle = new MinBoundingRectangle(dim);

            for(int i=0; i < dim; i++)
            {    
                overlapBoundingRectangle.setIntervalA(i, Math.max(getIntervalA(i), minBoundingRectangle.getIntervalA(i)));
                overlapBoundingRectangle.setIntervalB(i, Math.min(getIntervalB(i), minBoundingRectangle.getIntervalB(i)));
            }

            return overlapBoundingRectangle.getArea();
        }
        else {
            return 0;
        }
    }
    
    boolean isDimCutBy(int dim, double cut) {
        return intervals.get(dim).isDimCutBy(dim, cut);
    }

    public boolean isPrecise() {
        for(int i=0; i < dim; i++) {    
            if(getIntervalB(i) != getIntervalA(i)) {
                return false;
            }
        }
        
        return true;
    }
    
    public String toFullString()
    {
        String intervalsString = "{";
        for(int i = 0; i < dim; i++) {
            intervalsString += intervals.get(i).toFullString();
            
            if(i < dim - 1) {
                intervalsString += ", ";
            }
        }
        
        intervalsString += "}";
        
        return intervalsString;
    }
    
    public MinBoundingRectangle deepClone()
    {
        MinBoundingRectangle clone = new MinBoundingRectangle(dim);
        
        for(int i = 0; i < dim; i++) {
            clone.setInterval(i, getInterval(i).deepClone());
        }
        
        return clone;
    }
    
    private List<Node> nodes = new ArrayList<>();
    public List<Node> getNodes() {
        return nodes;
    }

    public void setNodes(List<Node> nodes) {
        this.nodes = nodes;
    }
    
    public void pickPreciseNodes(List<Node> nodes) {
        this.nodes.clear();
        
        for(Node node:nodes) {
            if(node.getMinBoundingRectangle().isPrecise() &&
               this.isOverlapped(node.getMinBoundingRectangle())) {
                this.nodes.add(node);
            }
        }
    }
    
    public void pickPreciseNodes(Node node, int dim) {
        for(MinBoundingRectangle box:children) {
            if(node.getMinBoundingRectangle().isPrecise() &&
               box.isOverlapped(node.getMinBoundingRectangle())) {
                if(dim == getDim())
                {
                    this.nodes.add(node);
                }
                else 
                {
                    box.pickPreciseNodes(node, dim + 1);
                }
            }
        }
    }

    public void pickPreciseNodes(List<Node> nodes, int dim) {
        for(MinBoundingRectangle box:children) {
            for(Node node:nodes) {
                if(node.getMinBoundingRectangle().isPrecise() &&
                   box.isOverlapped(node.getMinBoundingRectangle())) {
                    if(dim == getDim())
                    {
                        this.nodes.add(node);
                    }
                    else 
                    {
                        box.pickPreciseNodes(node, dim + 1);
                    }
                }
            }
        }
    }

    private List<MinBoundingRectangle> neighbour = new ArrayList<>();
    public List<MinBoundingRectangle> getNeighbours() {
        return neighbour;
    }
    public void setNeighbours(List<MinBoundingRectangle> neighbours) {
        this.neighbour = neighbour;
    }

    private List<MinBoundingRectangle> children = new ArrayList<>();
    public List<MinBoundingRectangle> getChildren() {
        return children;
    }
    public void setChildren(List<MinBoundingRectangle> children) {
        this.children = children;
    }

    public void pickNeighbour(List<MinBoundingRectangle> minBoundingRectangles, double neighbourMaxDistance) {
        this.neighbour.clear();
        
        for(MinBoundingRectangle minBoundingRectangle:minBoundingRectangles) {
            if(getCenterDistance(minBoundingRectangle) <=  neighbourMaxDistance) {
                this.neighbour.add(minBoundingRectangle);
            }
        }
    }
    
    public void assignNeighbour(MinBoundingRectangle minBoundingRectangle, double neighbourMaxDistance) {
        this.neighbour.clear();
        
        for(MinBoundingRectangle child:getChildren()) {
            if(child.getChildren().size() > 0) {
               child.assignNeighbour(minBoundingRectangle, neighbourMaxDistance);
            }
            else
            {
                child.pickNeighbour(minBoundingRectangle, neighbourMaxDistance);
                child.calculateKnnDensity();
            }
        }
    }
    
    public double density;
    public double getDensity() {
        return density;
    }

    public void setDensity(double density) {
        this.density = density;
    }
    
    public void calculateKnnDensity() {
        double sum = 0;
        for(MinBoundingRectangle neighbourNode:neighbour) {
            sum =+ neighbourNode.getNodes().size() / neighbourNode.getArea() / neighbourNode.getCenterDistance(this);
        }
        
        density = sum;
    }

    public void pickNeighbour(MinBoundingRectangle minBoundingRectangle, double neighbourMaxDistance) {
        this.neighbour.clear();
        
        for(MinBoundingRectangle child:minBoundingRectangle.getChildren()) {
            if(getCenterDistance(child) < neighbourMaxDistance) {
                if(child.getChildren().size() > 0) {
                   this.pickNeighbour(child, neighbourMaxDistance);
                }
                else
                {
                    this.neighbour.add(child);
                }
            }
        }
    }

    public double getCenterDistance(MinBoundingRectangle minBoundingRectangle) {
        double sum = 0;

        for(int i = 0; i < dim; i++)
        {
            sum += Math.pow(minBoundingRectangle.getIntervals().get(i).getCenter() - intervals.get(i).getCenter(), 2);
        }
            
        sum += Math.pow(sum, .5);
        
        return sum;
    }

    public double getPreciseCenterDistance(MinBoundingRectangle minBoundingRectangle) {
        double sum = 0;

        for(int i = 0; i < dim; i++)
        {
            sum += Math.pow(minBoundingRectangle.getIntervals().get(i).getPreciseCenter() - intervals.get(i).getPreciseCenter(), 2);
        }
            
        sum += Math.pow(sum, .5);
        
        return sum;
    }

    public double getCenterDistanceToParentMaxPoint() {
        double sum = 0;

        for(int i = 0; i < dim; i++)
        {
            sum += Math.pow(getParent().getIntervalB(i) - intervals.get(i).getCenter(), 2);
        }
            
        sum += Math.pow(sum, .5);
        
        return sum;
    }
    
    public void divide(double avgUnitSize, int dim) {
        children.clear();
        
        int n = (int)Math.round((getInterval(dim).getB() - getInterval(dim).getA()) / avgUnitSize + .5);
        
        for(int i = 0; i < n; i++) {
            MinBoundingRectangle box = new MinBoundingRectangle(getDim());
            box.setIntervalA(dim, getIntervalA(dim) + avgUnitSize * i);
            box.setIntervalB(dim, getIntervalA(dim) + avgUnitSize * (i + 1));
            
            for(int dim_i = 0; dim_i < getDim(); dim_i++) {
                if(dim_i != dim) {
                    box.setIntervalA(dim_i, getIntervalA(dim_i));
                    box.setIntervalB(dim_i, getIntervalB(dim_i));
                }
            }
            
            if(dim + 1 < getDim()) {
                box.divide(avgUnitSize, dim + 1);
            }
            
            children.add(box);
        }
    }
    
    Builder builder;
    public Builder getBuilder() {
        if(builder == null) {
            builder = new Builder();
        }
        return builder;
    }
}
