/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab;

import spatial.tree.lab.utility.ByteBufferUtility;
import spatial.tree.lab.utility.FileUtility;
import java.io.IOException;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.List;
import spatial.tree.lab.Interval.Limit;
import spatial.tree.lab.utility.Rankable;

/**
 *
 * @author Lawal 
 */
public class Node implements Rankable {

    public class Builder {
        Builder setDim(int dim) {
            Node.this.createMinBoundingRectangle(dim);
            Node.this.createSplittingHistory(dim);

            Node.this.dim = dim;
            return this;
        }

        Builder setMinEntries(int minEntries) {
            Node.this.minEntries = minEntries;
            return this;
        }

        Builder setMaxEntries(int maxEntries) {
            Node.this.maxEntries = maxEntries;
            return this;
        }
        
        Builder setIsData(boolean isData) {
            Node.this.isData = isData;
            return this;
        }

        Builder setIsSelected(boolean isSelected) {
            Node.this.isSelected = isSelected;
            
            if(parent != null) parent.getBuilder().setIsSelected(isSelected);
            return this;
        }

        Builder setIsSuperNode(boolean isSuperNode) {
            Node.this.isSuperNode = isSuperNode;
            return this;
        }

        Builder setIsSuperNode() {
            Node.this.isSuperNode = true;
            return this;
        }

        Builder setIsPacked(boolean isPacked) {
            Node.this.isPacked = isPacked;
            return this;
        }

        Builder setIsPacked() {
            Node.this.isPacked = true;
            return this;
        }

        public Builder setId(String id) {
            Node.this.id = id;
            return this;
        }
        
        public Builder setParent(Node parent) {
            Node.this.parent = parent;
            return this;
        }

        public Builder setMinBoundingRectangle(MinBoundingRectangle minBoundingRectangle) {
            Node.this.minBoundingRectangle = minBoundingRectangle;
            return this;
        }

        public Builder setInterval(int dim, double a, double b) {
            Node.this.setIntervalA(dim, a);
            Node.this.setIntervalB(dim, b);
            return this;
        }

        public Builder setInterval(int dim, double v) {
            Node.this.setIntervalA(dim, v);
            Node.this.setIntervalB(dim, v);
            return this;
        }
        
        Builder addEntryTemporary(Node entry) {
            Node.this.getEntries().add(entry);
            Node.this.updateMinBoundingRectangle();
            return this;
        }

        Builder addEntry(Node entry) {
            Node.this.getEntries().add(entry);
            entry.setParent(Node.this);
            Node.this.updateMinBoundingRectangle();
            return this;
        }

        Builder addEntryNoUpdate(Node entry) {
            Node.this.getEntries().add(entry);
            entry.setParent(Node.this);
            return this;
        }

        Builder addEntriesTemporary(List<Node> entries) {
            for(Node entry:entries) {
                addEntryTemporary(entry);
            }
            return this;
        }

        Builder addEntries(List<Node> entries) {
            for(Node entry:entries) {
                addEntry(entry);
            }
            return this;
        }

        Builder removeEntry(Node entry) {
            Node.this.getEntries().remove(entry);
            Node.this.updateMinBoundingRectangle();
            return this;
        }

        Builder removeEntryNoUpdate(Node entry) {
            Node.this.getEntries().remove(entry);
            Node.this.updateMinBoundingRectangle();
            return this;
        }

        Builder removeEntries(List<Node> entries) {
            for(Node entry:entries) {
                removeEntry(entry);
            }
            return this;
        }
        
        Builder setIsLoaded(boolean isLoaded) {
            Node.this.isLoaded = isLoaded;
            return this;
        }

        Builder  setFile(RandomAccessFile file) {
            Node.this.file = file;
            return this;
        }
        
        Builder  setDebugMode(int debugMode) {
            Node.this.debugMode = debugMode;
            return this;
        }

        public Node build() {
            return Node.this;
        }
    }
    
    int hashCode;
    
    int sequence = -1;
    public int getSequence() {
        return sequence;
    }
    public void setSequence(int sequence) {
        this.sequence = sequence;
    }
    
    int sequenceRange = 1;
    public int getSequenceRange() {
        return sequenceRange;
    }
    void setSequenceRange(int sequenceRange) {
        this.sequenceRange = sequenceRange;
    }
    
    String id;
    public String getId() {
        return id;
    }
    private void setId(String id) {
        this.id = id;
    }
    
    int dim;
    public int getDim() {
        return dim;
    }
    private void setDim(int dim) {
        this.dim = dim;
    }
    
    int minEntries;
    public int getMinEntries() {
        return minEntries;
    }
    private void setMinEntries(int minEntries) {
        this.minEntries = minEntries;
    }

    int maxEntries;
    public int getMaxEntries() {
        return maxEntries;
    }
    private void setMaxEntries(int maxEntries) {
        this.maxEntries = maxEntries;
    }
        
    Node parent;
    public Node getParent() {
        return parent;
    }
    private void setParent(Node parent) {
        this.parent = parent;
    }
    
    boolean isData;
    boolean getIsData() {
        return isData;
    }
    private void setIsData(boolean isData) {
        this.isData = isData;
    }
    
    public void loadNode() {
        if(sequence > -1 && !isLoaded) {
            FileUtility fileUtility = new FileUtility();
            
            try
            {
                byte[] bytes = fileUtility.getBytesFromRandomAccessFile(file, (long)((long)sequence * (long)ByteBufferUtility.MEMORY_PAGE_SIZE), sequenceRange * ByteBufferUtility.MEMORY_PAGE_SIZE);
                fromByteArray(bytes);
            }
            catch(Exception e) {
                System.out.println(e.getMessage());
            }
        }
    }
    
    private List<Node> children = new ArrayList<>();
    public List<Node> getChildren() {
        return children;
    }
    public void setChildren(List<Node> children) {
        this.children = children;
    }

    List<Node> entries = new ArrayList<>();
    public List<Node> getEntries() {
        loadNode();
        return entries;
    }
    private void setEntries(List<Node> entries) {
        this.entries = entries;
    }
     
    MinBoundingRectangle minBoundingRectangle;
    public MinBoundingRectangle getMinBoundingRectangle() {
        return minBoundingRectangle;
    }
    private void setMinBoundingRectangle(MinBoundingRectangle minBoundingRectangle) {
        this.minBoundingRectangle = minBoundingRectangle;
    }
    
    boolean isSuperNode;
    boolean isIsSuperNode() {
        return isSuperNode;
    }
    private void setIsSuperNode(boolean isSuperNode) {
        this.isSuperNode = isSuperNode;
    }

    boolean isPacked;
    boolean isIsPacked() {
        return isPacked;
    }
    private void setIsPacked(boolean isPacked) {
        this.isPacked = isPacked;
    }

    List<Boolean> splitDimHistory;
    public List<Boolean> getSplitDimHistory() {
        return splitDimHistory;
    }
    private void setSplitDimHistory(List<Boolean> splitDimHistory) {
        this.splitDimHistory = splitDimHistory;
    }
    
    boolean isSelected;
    public boolean getIsSelected() {
        return isSelected;
    }

    void setIsSelected(boolean isSelected) {
        this.isSelected = isSelected;
    }
    
    
    RandomAccessFile file;
    RandomAccessFile getFile() {
        return file;
    }
    void setFile(RandomAccessFile file) {
        this.file = file;
    }
    
    public Node(){
        hashCode = ((Object)this).hashCode();
    }

    Builder builder;
    public Builder getBuilder() {
        if(builder == null) {
            builder = new Builder();
        }
        return builder;
    }

    void createMinBoundingRectangle(int dim)
    {
        if( minBoundingRectangle == null ) {
            minBoundingRectangle = new MinBoundingRectangle(dim);
        }
    }
    
    void createSplittingHistory(int dim) {
        if(splitDimHistory == null ) {
            splitDimHistory = new ArrayList<> ();
            for(int i = 0; i < dim; i++) {
                splitDimHistory.add(false);
            }
        }
    }

    void splittedByDim(int dim) {
            splitDimHistory.set(dim, true);
    }

    public Node getEntry(int i) {
        return getEntries().get(i);
    }

    public int getEntriesCount()
    {
        return getEntries().size();
    }
    
    int debugMode = 0;
    public int getDebugMode() {
        return debugMode;
    }
    void setDebugMode(int debugMode) {
        this.debugMode = debugMode;
    }
    
    public String toFullString(int level) {
        
        String name = parent == null ? "" : ((Object)parent).hashCode() + ">";
        name += ((Object)this).hashCode() +  " ";
        
        if(debugMode == 0) {
            name = "";
        }
        
        if(id == null && this.parent != null) {
            name += "Level " + Integer.toString(level) + 
                    " Entry " + Integer.toString(this.parent.getEntries().indexOf(this) + 1);
        }
        else {
            name += " " + id;
        }
        String levelIdent = "";
        for(int i =0; i < level; i++) {
            levelIdent += "\t";
        }
        
        String addtions = "";
        if(isSuperNode) addtions +="S";
        if(isPacked) addtions +="P";
        
        if(addtions.length() > 0) addtions = " (" + addtions + ")";
        
        return  levelIdent + " " + name + addtions + (minBoundingRectangle == null ? " " : " " + minBoundingRectangle.toFullString());
    }

    public String toFullString() {
        return toFullString(0);
    }

    boolean ExceedMaxEntries() {
        return getEntries().size() > maxEntries;
    }

    boolean isRoot() {
        return parent == null;
    }
    
    boolean isLeaf() {
        return isData;
    }

    public boolean isLeafNode() {
        return (getEntriesCount() == 0 && !isData) || 
               (getEntriesCount() > 0 && getEntries().get(0).isLeaf());
    }

    boolean isNonLeafNode() {
        return !isData && !isLeafNode();
    }

    public int getLevel() {
        if(isRoot()) {
            return 0;
        }
        else {
            int level = 1;
            Node lastParent = parent;
            while(!lastParent.isRoot()) {
                lastParent = lastParent.parent;
                level++;
            }
            
            return level;
        }
    }
    
    public int getLastLevel() {
        if(getEntries().size() > 0) {
            return getEntries().get(0).getLastLevel();
        }
        else {
            return getLevel();
        }
    }
    
    public Interval getInterval(int dim) {
        return minBoundingRectangle.getInterval(dim);
    }

    public double getIntervalA(int dim) {
        return minBoundingRectangle.getIntervalA(dim);
    }
    
    public double getIntervalC(int dim) {
        return minBoundingRectangle.getIntervalC(dim);
    }

    public double getIntervalA(int dim, int limit) {
        if(limit == 0)
        {
            return minBoundingRectangle.getIntervalA(dim);
        }
        else
        {
            return minBoundingRectangle.getIntervalB(dim);
        }
    }

    void setIntervalA(int dim, double a) {
        minBoundingRectangle.setIntervalA(dim, a);
    }

    public double getIntervalB(int dim) {
        return minBoundingRectangle.getIntervalB(dim);
    }
    
    public double getIntervalLimit(int dim, Limit limit) 
    {
        if(limit == Limit.LOWER)
        {
            return getIntervalA(dim);
        }
        else
        {
            return getIntervalB(dim);
        }
    }

    public double getIntervalCenter(int dim) 
    {
        return (getIntervalA(dim) + getIntervalB(dim)) / 2;
    }

    void setIntervalB(int dim, double b) {
        minBoundingRectangle.setIntervalB(dim, b);
    }

    void updateMinBoundingRectangle() {
        minBoundingRectangle.setForAllAB(Double.POSITIVE_INFINITY, Double.NEGATIVE_INFINITY);
        
        for(int i = 0; i < dim; i++)
        {
            for (Node entry : getEntries()) {
                minBoundingRectangle.getInterval(i).enlarge(entry.getMinBoundingRectangle().getInterval(i));
            }
        }
        
        if(parent != null) {
            parent.updateMinBoundingRectangle();
        }
    }

    List<Double> sortRank = new ArrayList<>();
    public List<Double> getSortRank() {
        return sortRank;
    }
    public void setRank(List<Double> sortRank) {
        this.sortRank = sortRank;
    }
    
    public double getEnlargement(Node entry) {
        return this.getBuilder().addEntryTemporary(entry).build().getArea() - 
                this.getBuilder().removeEntry(entry).build().getArea();
    }
    
    public double getArea() {
        return minBoundingRectangle.getArea();
    }
    
    public double getMargin() {
        return minBoundingRectangle.getMargin();
    }

    public boolean isOverlapped(Node node) {
        return this.minBoundingRectangle.isOverlapped(node.getMinBoundingRectangle());
    }
    
    public double getOverlap(Node node) {
        return minBoundingRectangle.getOverlapArea(node.getMinBoundingRectangle());
    }

    public double getOverlapPercent(Node node) {
        return getOverlap(node) / (getArea() + node.getArea() -  getOverlap(node));
    }
    
    public double getCenterDistance(Node node) {
        double sum = 0;
        for (Node entry : getEntries()) {
            double entryDiatance = 0;
            
            for(int i = 0; i < dim; i++)
            {
                entryDiatance += Math.pow(entry.getIntervalCenter(i) - getIntervalCenter(i), 2);
            }
            
            sum += Math.pow(entryDiatance, .5);
        }
        
        return sum;
    }
    
    boolean isDimCutBy(int dim, double cut) {
        return minBoundingRectangle.isDimCutBy(dim, cut);
    }
    
    void PackInside(Node parentNode) {
        for(int i = 0; i < dim; i++) {
            if(minBoundingRectangle.getIntervalA(i) < parentNode.getIntervalA(i)){
               minBoundingRectangle.setIntervalA(i, parentNode.getIntervalA(i));
               isPacked = true;
            }
            if(minBoundingRectangle.getIntervalB(i) > parentNode.getIntervalB(i)){
               minBoundingRectangle.setIntervalB(i, parentNode.getIntervalB(i));
               isPacked = true;
            }
        }
    }
    
    public Node deepClone()
    {
        Node clone = new Node();
        clone.getBuilder().
            setDim(dim).
            setId(id).
            setMinEntries(minEntries).
            setMaxEntries(maxEntries).
            setParent(parent).
            setIsData(isData).
            setIsSuperNode(isSuperNode).
            setIsPacked(isPacked).
            setMinBoundingRectangle(minBoundingRectangle.deepClone()).
            setDebugMode(debugMode);
        
        for(Node entry:entries) {
            clone.getBuilder().addEntry(entry.deepClone());
        }
        
        return clone;
    }
    
    void allocateSeuence(RTreeBase rTree) {
        if(sequence == -1) {
            setSequence(rTree.getNewSequence(sequenceRange));
            sequenceRange = getPagesCount();
        }
        else {
            if(isLoaded) {
                if(getPagesCount() > sequenceRange) {
                    setSequence(rTree.getNewSequence(sequenceRange));
                    sequenceRange = getPagesCount();
                }
            }
        }
    }

    void deepAllocateSequence(RTreeBase rTree) {
        allocateSeuence(rTree);
        
        for(Node entry: getEntries()) {
            if(!entry.isData) {
                entry.deepAllocateSequence(rTree);
            }
        }
    }
    
    boolean isLoaded = true;
    public boolean getIsLoaded() {
        return isLoaded;
    }
    public void setIsLoaded(boolean isLoaded) {
        this.isLoaded = isLoaded;
    }
    
    public int getMemorySize() {
        int size = 0;

        size = size + 32 * 2;  //this.id

        size = size + Integer.BYTES; //this.dim
        size = size + Integer.BYTES; //this.minEntries
        size = size + Integer.BYTES; //this.maxEntries

        size = size + 1; //this.isData;
        size = size + 1; //this.isPacked;
        size = size + 1; //this.isSuperNode;
        
        size = size + this.splitDimHistory.size(); //item
        
        size = size + Double.BYTES * this.dim * 2; //this.minBoundingRectangle;
        
        size = size + Integer.BYTES; //this.sequence;
        
        size = size + Integer.BYTES; //this.sequenceRange;

        size = size + Integer.BYTES; //this.getEntriesCount()
        
        size = size + Integer.BYTES * this.getEntriesCount(); //entry.sequence;
        size = size + Integer.BYTES * this.getEntriesCount(); //entry.sequenceRange;
        size = size + 32 * this.getEntriesCount(); //this.entry.id;
        size = size + this.getEntriesCount(); //entry.isData;
        size = size + Double.BYTES * this.getEntriesCount() * this.dim * 2; //entry.minBoundingRectangle;

        return size;
    }

    public int getMaxNodeMemorySize() {
        int size = 0;

        size = size + 32 * 2;  //this.id

        size = size + Integer.BYTES; //this.dim
        size = size + Integer.BYTES; //this.minEntries
        size = size + Integer.BYTES; //this.maxEntries

        size = size + 1; //this.isData;
        size = size + 1; //this.isPacked;
        size = size + 1; //this.isSuperNode;
        
        size = size + this.splitDimHistory.size(); //item
        
        size = size + Double.BYTES * this.dim * 2; //this.minBoundingRectangle;
        
        size = size + Integer.BYTES; //this.sequence;
        size = size + Integer.BYTES; //this.sequenceRange;

        size = size + Integer.BYTES; //this.getEntriesCount()
        
        size = size + Integer.BYTES * this.maxEntries; //entry.sequence;
        size = size + Integer.BYTES * this.maxEntries; //entry.sequenceRange;
        size = size + 32 * this.maxEntries; //this.entry.id;
        size = size + this.maxEntries; //entry.isData;
        size = size + Double.BYTES * this.maxEntries * this.dim * 2; //entry.minBoundingRectangle;

        return size;
    }

    public long deepGetMemorySize() {
        long  total = getMemorySize();
        
        for(Node entry:getEntries()) {
            if(!entry.isData) {
                 total = total + (long)entry.deepGetMemorySize();
            }
        }
        
        return total;
    }

    public byte[] toByteArray() {
        ByteBufferUtility bb = new ByteBufferUtility(ByteBufferUtility.MEMORY_PAGE_SIZE * getPagesCount());
        
        //this.id
        bb.addString(String.format("%1$-16s", id == null ? "" : id));
        
        //this.dim
        bb.addInteger(dim);
        //this.minEntries
        bb.addInteger(minEntries);
        //this.maxEntries
        bb.addInteger(maxEntries);

        //this.isData
        bb.addBoolean(isData);
        //this.isPacked
        bb.addBoolean(isPacked);
        //this.isSuperNode
        bb.addBoolean(isSuperNode);
        
        //this.splitDimHistory
        for(boolean item:splitDimHistory) {
            bb.addBoolean(item);
        }

        for(int i = 0; i < dim; i++) {
            //this.minBoundingRectangle.getIntervalA
            bb.addDouble(minBoundingRectangle.getIntervalA(i)); 

            //this.minBoundingRectangle.getIntervalB
            bb.addDouble(minBoundingRectangle.getIntervalB(i)); 
        }

        //this.sequence;
        bb.addInteger(sequence);

        //this.sequenceRange;
        bb.addInteger(sequenceRange);

        //this.getEntriesCount()
        bb.addInteger(getEntriesCount());
        
        for(Node entry:getEntries()) {
            //entry.sequence
            bb.addInteger(entry.sequence);
            
            //entry.sequenceRange
            bb.addInteger(entry.sequenceRange);
            
            //entry.id
            bb.addString(String.format("%1$-16s", entry.id == null ? "" : entry.id)); 
            
            //entry.isData
            bb.addBoolean(entry.isData);
            
            for(int i = 0; i < dim; i++) {
                //entry.getIntervalA
                bb.addDouble(entry.getIntervalA(i)); 
                
                //entry.getIntervalB
                bb.addDouble(entry.getIntervalB(i)); 
            }
        }

        return bb.getBuffer().array();
    }
    
    public void fromByteArray(byte[] bytes) {
        ByteBufferUtility bb = new ByteBufferUtility(bytes.length);
        bb.getBuffer().put(bytes);
        bb.getBuffer().flip();
        
        //this.id
        id = bb.getString();
        id = id.trim().equals("") ? null : id.trim();
        
        //this.dim
        dim = bb.getInteger();
        //this.minEntries
        minEntries = bb.getInteger();
        //this.maxEntries
        maxEntries = bb.getInteger();

        //this.isData
        isData = bb.getBoolean();
        //this.isPacked
        isPacked = bb.getBoolean();
        //this.isSuperNode
        isSuperNode = bb.getBoolean();
        
        //this.splitDimHistory
        splitDimHistory = null;
        createSplittingHistory(dim);
        
        for(int i = 0; i < dim; i++) {
            splitDimHistory.set(i, bb.getBoolean());
        }

        // this.minBoundingRectangle
        minBoundingRectangle = null;
        createMinBoundingRectangle(dim);
        for(int i = 0; i < this.dim; i++) {
            //this.minBoundingRectangle.setIntervalA
            //double a = bb.getDouble();
            minBoundingRectangle.setIntervalA(i, bb.getDouble());

            //this.minBoundingRectangle.setIntervalB
            //double b = bb.getDouble();
            minBoundingRectangle.setIntervalB(i, bb.getDouble()); 
        }
        
        //this.sequence;
        setSequence(bb.getInteger());

        //this.sequenceRange;
        sequenceRange = bb.getInteger();

        if(!isData) {
            //this.getEntriesCount()
           int entriesCount = bb.getInteger();

            for(int e = 0; e < entriesCount; e++) {
                Node entry = new Node();
                entry.file = file;
                
                entry.isLoaded = false;
                entries.add(entry);
                entry.setParent(this);
                
                //entry.sequence
                entry.setSequence(bb.getInteger());
                
                //entry.sequenceRange
                entry.setSequenceRange(bb.getInteger());

                //entry.id
                entry.id = bb.getString();
                entry.id = entry.id.equals("") ? null : entry.id;
                
                //entry.isDaya
                entry.isData = bb.getBoolean();
                
                // this.minBoundingRectangle
                entry.minBoundingRectangle = null;
                entry.createMinBoundingRectangle(dim);
                
                for(int i = 0; i < this.dim; i++) {
                    //entry.setIntervalA
                    //double a = bb.getDouble();
                    entry.setIntervalA(i, bb.getDouble());
                    
                    //entry.setIntervalB
                    //double b = bb.getDouble();
                    entry.setIntervalB(i, bb.getDouble()); 
                }
            }
        }
        
        isLoaded = true;
    }

    public void writeToByteArray(RandomAccessFile out) throws IOException {
        if(isLoaded && sequence > -1) {
            byte[] source = toByteArray();

            if((long)((long)ByteBufferUtility.MEMORY_PAGE_SIZE * (long)sequence) < 0) {
                System.out.println("Negative");
            }
            
            out.seek((long)((long)ByteBufferUtility.MEMORY_PAGE_SIZE * (long)sequence));
            
            out.write(source);

            byte[] fill = new byte[ByteBufferUtility.MEMORY_PAGE_SIZE * getPagesCount() - source.length];
            out.write(fill);
        }
    }

    public void deepWriteToByteArray(RandomAccessFile out) throws IOException {
        if(isLoaded && sequence > -1) {
            writeToByteArray(out);

            for(Node entry:getEntries()) {
                if(!entry.isData) {
                    entry.deepWriteToByteArray(out);
                }
            }
        }
    }

    public int getPagesCount() {
        int count = (getEntriesCount() / getMaxEntries() + (getEntriesCount() % getMaxEntries() == 0 ? 0 : 1));
        if(count == 0) count = 1;
        return count;
    } 
    
    public long getHardDriveSize() {
        if(isData) {
            return 0;
        }
        else {
            return ByteBufferUtility.MEMORY_PAGE_SIZE * getPagesCount();
        }
    }

    public long deepGetHardDriveSize() {
        long total = getHardDriveSize();
        
        for(Node entry:getEntries()) {
            if(!isData) {
                total = total + entry.deepGetHardDriveSize();
            }
        }
        
        return total;
    }

    public long deepGetNodesCount() {
        long result = 1;
         
        for(Node entry:getEntries()) {
            if(!entry.isData) {
                result += entry.deepGetNodesCount();
            }
        }
        
        return result;
     }

    public int deepGetLeavesCount() {
        if(isData) return 1;
        
        int result = 0;
         
        for(Node entry:getEntries()) {
            result += entry.deepGetLeavesCount();
        }
        
        return result;
    }
        
    public RQueryResult getNodes(RTreeBase rtree, MinBoundingRectangle minBoundingRectangle) {
        RQueryResult result = new RQueryResult();
        if(!isData) 
        {
            result.getVisitedNodes().add(this);

            for(Node entry:getEntries()) {
                if(entry.minBoundingRectangle.isOverlapped(minBoundingRectangle)) {
                    if(entry.isData) {
                        result.getLeaves().add(entry);
                    }
                    else {
                        result.appendNodeResult(entry.getNodes(rtree, minBoundingRectangle));
                    }
                }
            }
        }
         
        return result;
    }
    
    public RQueryResult getNodes(RTreeBase rtree, int d, Interval interval) {
        RQueryResult result = new RQueryResult();
        if(!isData) 
        {
            result.getVisitedNodes().add(this);

            for(Node entry:getEntries()) {
                if(entry.getInterval(d).isOverlapped(interval)) {
                    if(entry.isData) {
                        result.getLeaves().add(entry);
                    }
                    else {
                        result.appendNodeResult(entry.getNodes(rtree, d, interval));
                    }
                }
            }
        }
         
        return result;
    }

    private List<MinBoundingRectangle> neighbour = new ArrayList<>();
    public List<MinBoundingRectangle> getNeighbours() {
        return neighbour;
    }
    public void setNeighbours(List<MinBoundingRectangle> neighbours) {
        this.neighbour = neighbour;
    }
    
    public int getImpreciseDim() {
        for(int i = 0; i < getMinBoundingRectangle().getDim(); i++) {
            if(getMinBoundingRectangle().getInterval(i).getA() != getMinBoundingRectangle().getInterval(i).getB()) {
                return i;
            }
        }
        
        return -1;
    }

    public void pickNeighbour(MinBoundingRectangle mainMinBoundingRectangle) {
        this.neighbour.clear();
        
        for(MinBoundingRectangle child:mainMinBoundingRectangle.getChildren()) {
            if(minBoundingRectangle.isOverlapped(child)) {
                if(child.getChildren().size() > 0) {
                   this.pickNeighbour(child);
                }
                else
                {
                    this.neighbour.add(child);
                }
            }
        }
    }
    
    public void pocessEntries() {
        for(Node entry:this.getEntries()) {
            entry.setParent(this);
        }
    }

    void assertEntries() {
    }
}