/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab;

import spatial.tree.lab.Interval.Limit;


/**
 *
 * @author Lawal
 */
class RAsteriskChooseSplitAxisAlgo {
    class Input {
        Node node;
        Node getNode() {
            return node;
        }
        void setNode(Node node) {
            this.node = node;
        }
    }

    static class Output {
        int axis;
        int getAxis() {
            return axis;
        }
        void setAxis(int axis) {
            this.axis = axis;
        }

        Node bestOverlapNode1;
        Node getBestOverlapNode1() {
            return bestOverlapNode1;
        }
        void setBestOverlapNode1(Node bestOverlapNode1) {
            this.bestOverlapNode1 = bestOverlapNode1;
        }
        
        Node bestOverlapNode2;
        Node getBestOverlapNode2() {
            return bestOverlapNode2;
        }
        void setBestOverlapNode2(Node bestOverlapNode2) {
            this.bestOverlapNode2 = bestOverlapNode2;
        }
    }

    class Builder {
        Builder setNode(Node node) {
            RAsteriskChooseSplitAxisAlgo.this.input.node = node;
            return this;
        }

        RAsteriskChooseSplitAxisAlgo build() {
            return RAsteriskChooseSplitAxisAlgo.this;
        }
    }

    Input input = new Input();
    Input getInput() {
        return input;
    }
    void setInput(Input input) {
        this.input = input;
    }
    
    Output output = new Output();
    Output getOutput() {
        return output;
    }
    void setOutput(Output output) {
        this.output = output;
    }
    
    Builder builder;
    public Builder getBuilder() {
        if(builder == null) {
            builder = new Builder();
        }
        return builder;
    }

    void execute(RTreeBase rTree) {
        if(rTree.getDebugMode() >= 1) {
            System.out.println("Choose Split Axis Algo Input");
            System.out.println("Node:");
            rTree.deepPrint(input.node);
            System.out.println();
        }

        Node node = rTree.createIntermediateNode();
        
        node.getBuilder().addEntriesTemporary(input.node.getEntries());
        
        double bestS = Double.POSITIVE_INFINITY;
        double bestOverlap = Double.POSITIVE_INFINITY;

        for(Limit limit: Limit.values()) {
            for(int i = 0; i < input.node.getDim(); i++) {
                node.getEntries().sort(new UtilitySortEntries(i, limit, UtilitySortEntries.SortEntriesBy.LIMIT));

                for(int j = 0; j < input.node.getMaxEntries() - 2 * input.node.getMinEntries() + 2; j++) {
                    
                    Node node1 = rTree.createIntermediateNode();
                    for(int k = 0; k < input.node.getMinEntries() + j; k++) {
                        node1.getBuilder().addEntryTemporary(input.node.getEntry(k));
                    }
                    
                    Node node2 = rTree.createIntermediateNode();
                    for(int k = input.node.getMinEntries() + j; k < input.node.getEntriesCount(); k++) {
                        node2.getBuilder().addEntryTemporary(input.node.getEntry(k));
                    }
                                    
                    double currentS = node1.getMargin() + node2.getMargin();

                    if(bestS > currentS)
                    {
                        bestS = currentS;

                        output.axis = i / 2;
                    }
                    
                    if(bestOverlap > node1.getOverlap(node2)) {
                        output.bestOverlapNode1 = node1;
                        output.bestOverlapNode2 = node2;
                    }

                    if(bestOverlap == node1.getOverlap(node2) &&
                        Math.abs(output.bestOverlapNode1.getEntriesCount() - output.bestOverlapNode2.getEntriesCount()) > 
                        Math.abs(node1.getEntriesCount() - node2.getEntriesCount())) {
                        output.bestOverlapNode1 = node1;
                        output.bestOverlapNode2 = node2;
                    }
                }
            }
        }

        if(rTree.getDebugMode() >= 1) {
            System.out.println("Choose Split Axis Algo Output");
            System.out.println("BestOverlapNode1:");
            rTree.deepPrint(output.bestOverlapNode1);
            System.out.println("BestOverlapNode2:");
            rTree.deepPrint(output.bestOverlapNode2);
            System.out.println("Axis:");
            System.out.println(output.axis);
            System.out.println();
        }
    }
}
