/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab;

import java.util.Collections;
import java.util.Comparator;
import spatial.tree.lab.Interval.Limit;

/**
 *
 * @author Lawal
 */
class RAsteriskChooseSplitIndexAlgo {
    class Input {
        Node node;
        Node getNode() {
            return node;
        }
        void setNode(Node node) {
            this.node = node;
        }

        int axis;
        int getAxis() {
            return axis;
        }
        void setAxis(int axis) {
            this.axis = axis;
        }
    }

    static class Output {
        Node node1;
        Node getNode1() {
            return node1;
        }

        void setNode1(Node node1) {
            this.node1 = node1;
        }

        Node node2;
        Node getNode2() {
            return node2;
        }
        void setNode2(Node node2) {
            this.node2 = node2;
        }
    }

    class Builder {
        Builder setNode(Node node) {
            RAsteriskChooseSplitIndexAlgo.this.input.node = node;
            return this;
        }
        
        Builder setAxis(int axis) {
            RAsteriskChooseSplitIndexAlgo.this.input.axis = axis;
            return this;
        }

        RAsteriskChooseSplitIndexAlgo build() {
            return RAsteriskChooseSplitIndexAlgo.this;
        }
    }

    Input input = new Input();
    Input getInput() {
        return input;
    }
    void setInput(Input input) {
        this.input = input;
    }
    
    Output output = new Output();
    Output getOutput() {
        return output;
    }
    void setOutput(Output output) {
        this.output = output;
    }
        
    Builder builder;
    public Builder getBuilder() {
        if(builder == null) {
            builder = new Builder();
        }
        return builder;
    }

    void execute(RTreeBase rTree) {
        if(rTree.getDebugMode() >= 1) {
            System.out.println("Choose Split Index Algo Input");
            System.out.println("Node:");
            rTree.deepPrint(input.node);
            System.out.println("Axis:");
            System.out.println(input.axis);
            System.out.println();
        }

        double bestOverlab = Double.POSITIVE_INFINITY;

        for(Limit limit: Limit.values()) {
            for(int i = 0; i < input.node.getDim(); i++) {
                Node node = rTree.createIntermediateNode();
                node.getBuilder().addEntriesTemporary(input.node.getEntries());
                
                node.getEntries().sort(new UtilitySortEntries(i, limit, UtilitySortEntries.SortEntriesBy.LIMIT));
                
                for(int j = 0; j < input.node.getMaxEntries() -2 * input.node.getMinEntries() + 2; j++)
                {
                    Node node1 = rTree.createIntermediateNode();
                    for(int k = 0; k < input.node.getMinEntries() + j; k++) {
                        node1.getBuilder().addEntryTemporary(input.node.getEntry(k));
                    }
                    
                    Node node2 = rTree.createIntermediateNode();
                    for(int k = input.node.getMinEntries() + j; k < input.node.getEntriesCount(); k++) {
                        node2.getBuilder().addEntryTemporary(input.node.getEntry(k));
                    }
                                    
                    double currentOverlab = node1.getMargin() + node2.getMargin();

                    if(bestOverlab > currentOverlab)
                    {
                        bestOverlab = currentOverlab;

                        output.node1 = node1; 
                        output.node2 = node2; 
                    }

                    if(bestOverlab == currentOverlab)
                    {
                        if(node1.getArea() + node2.getArea() < 
                                output.node1.getArea() + output.node2.getArea()) {
                            bestOverlab = currentOverlab;

                            output.node1 = node1;
                            output.node2 = node2;
                        }
                    }
                }
            }
        }
        
        if(rTree.getDebugMode() >= 1) {
            System.out.println("Choose Split Index Algo Output");
            System.out.println("Node1:");
            rTree.deepPrint(output.node1);
            System.out.println("Node2:");
            rTree.deepPrint(output.node2);
            System.out.println();
        }
    }
}
