/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab;

/**
 *
 * @author Lawal
 */
class RAsteriskChooseSubTreeAlgo {
    class Input {
        Node node;
        Node getNode() {
            return node;
        }
        void setNode(Node node) {
            this.node = node;
        }

        Node entry;
        Node getEntry() {
            return entry;
        }
        void setEntry(Node entry) {
            this.entry = entry;
        }

        int level;
        int getLevel() {
            return level;
        }
        void setLevel(int level) {
            this.level = level;
        }
    }

    static class Output {
        Node node;
        Node getNode() {
            return node;
        }
        void setNode(Node node) {
            this.node = node;
        }
    }

    class Builder {
        Builder setNode(Node node) {
            RAsteriskChooseSubTreeAlgo.this.input.node = node;
            return this;
        }

        Builder setEntry(Node entry) {
            RAsteriskChooseSubTreeAlgo.this.input.entry = entry;
            return this;
        }

        Builder setLevel(int level) {
            RAsteriskChooseSubTreeAlgo.this.input.level = level;
            return this;
        }
        Builder execute(RTreeBase rTree) {
            RAsteriskChooseSubTreeAlgo.this.execute(rTree);
            return this;
        }

        RAsteriskChooseSubTreeAlgo build() {
            return RAsteriskChooseSubTreeAlgo.this;
        }
    }

    Input input = new Input();
    Input getInput() {
        return input;
    }
    void setInput(Input input) {
        this.input = input;
    }
    
    Output output = new Output();
    Output getOutput() {
        return output;
    }
    void setOutput(Output output) {
        this.output = output;
    }

    Builder builder;
    public Builder getBuilder() {
        if(builder == null) {
            builder = new Builder();
        }
        return builder;
    }

    void execute(RTreeBase rTree) {
        if(rTree.getDebugMode() >= 1) {
            System.out.println("Choose Sub Tree Algo Input");
            System.out.println("Node:");
            rTree.deepPrint(input.node);
            System.out.println("Entry:");
            rTree.deepPrint(input.entry);
            System.out.println("Level:");
            System.out.println(input.level);
            System.out.println();
        }

        Node node = input.node;
        
        if(input.node.getLevel() == input.level)
        {
            output.node = node;
            return;
        }
        else
        {
            if(input.node.isLeafNode())
            {
                double bestOverlapEnlargment = Double.POSITIVE_INFINITY;
                Node bestNode = null;

                for(Node outerEntries:node.getEntries())
                {
                    double currentOverlapEnlargment = 0;
                    for(Node innerEntry:outerEntries.getEntries())
                    {
                        currentOverlapEnlargment += innerEntry.getOverlap(input.entry);
                    }
                    
                    if(currentOverlapEnlargment < bestOverlapEnlargment)
                    {
                        bestNode = outerEntries;
                        
                        bestOverlapEnlargment = currentOverlapEnlargment;
                    }
                    
                    if(currentOverlapEnlargment == bestOverlapEnlargment)
                    {
                        if(bestNode.getArea() > outerEntries.getArea())
                        {
                            bestNode = outerEntries;
                            bestOverlapEnlargment = currentOverlapEnlargment;
                        }
                    }
                }

                output.node = bestNode;
            }
            else
            {
                double bestEnlargment = Double.POSITIVE_INFINITY;
                Node bestNode = null;

                for(Node entry:node.getEntries())
                {
                    double currentEnlargment = entry.getEnlargement(input.entry);

                    if(currentEnlargment < bestEnlargment)
                    {
                        bestNode = entry;
                        bestEnlargment = currentEnlargment;
                    }

                    if(currentEnlargment == bestEnlargment)
                    {
                        if(bestNode.getArea() > entry.getArea())
                        {
                            bestNode = entry;
                            bestEnlargment = currentEnlargment;
                        }
                    }
                }

                output.node = bestNode;
            }
        }
        
        output.node = new RAsteriskChooseSubTreeAlgo().getBuilder().
                                                            setNode(output.node).
                                                            setEntry(input.entry).
                                                            setLevel(input.level).
                                                            execute(rTree).
                                                                build().
                                                                    output.node;

        if(rTree.getDebugMode() >= 1) {
            System.out.println("Choose Sub Tree Algo Output");
            System.out.println("Node:");
            rTree.deepPrint(output.node);
            System.out.println();
        }
    }
}
