/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab;

import java.util.ArrayList;
import java.util.List;

/**
 *
 * @author Lawal
 */
class RAsteriskOverflowTreatmentAlgo {
    class Input {
        Node node;
        Node getNode() {
            return node;
        }
        void setNode(Node node) {
            this.node = node;
        }
        
        int level;
        int getLevel() {
            return level;
        }
        void setLevel(int level) {
            this.level = level;
        }

        int reInsertCount;
        int getReInsertCount() {
            return reInsertCount;
        }
        void setReInsertCount(int reInsertCount) {
            this.reInsertCount = reInsertCount;
        }
    }

    class Output {
        Node node;
        Node getNode() {
            return node;
        }
        void setNode(Node node) {
            this.node = node;
        }
    }

    class Builder {
        Builder setNode(Node node) {
            RAsteriskOverflowTreatmentAlgo.this.input.node = node;
            return this;
        }

        Builder setLevel(int level) {
            RAsteriskOverflowTreatmentAlgo.this.input.level = level;
            return this;
        }

        Builder setReInsertCount(int reInsertCount) {
            RAsteriskOverflowTreatmentAlgo.this.input.reInsertCount = reInsertCount;
            return this;
        }

        RAsteriskOverflowTreatmentAlgo build() {
            return RAsteriskOverflowTreatmentAlgo.this;
        }
    }

    Input input = new Input();
    Input getInput() {
        return input;
    }
    void setInput(Input input) {
        this.input = input;
    }
    
    Output output = new Output();
    Output getOutput() {
        return output;
    }
    void setOutput(Output output) {
        this.output = output;
    }
    
    List<Integer> visitedLevels = new ArrayList<>();
    List<Integer> getVisitedLevels() {
        return visitedLevels;
    }
    void setVisitedLevels(List<Integer> visitedLevels) {
        this.visitedLevels = visitedLevels;
    }

    RAsteriskSplitNodeAlgo splitNodeAlgo = new RAsteriskSplitNodeAlgo();
    RAsteriskSplitNodeAlgo getSplitNodeAlgo() {
        return splitNodeAlgo;
    }
    void setSplitNodeAlgo(RAsteriskSplitNodeAlgo splitNodeAlgo) {
        this.splitNodeAlgo = splitNodeAlgo;
    }

    RAsteriskReInsertNodeAlgo reInsertNodeAlgo = new RAsteriskReInsertNodeAlgo();
    RAsteriskReInsertNodeAlgo getReInsertNodeAlgo() {
        return reInsertNodeAlgo;
    }
    void setReInsertNodeAlgo(RAsteriskReInsertNodeAlgo reInsertNodeAlgo) {
        this.reInsertNodeAlgo = reInsertNodeAlgo;
    }
    
    RAsteriskInsertNodeAlgo insertNodeAlgo;
    RAsteriskInsertNodeAlgo getInsertNodeAlgo() {
        return insertNodeAlgo;
    }
    void setInsertNodeAlgo(RAsteriskInsertNodeAlgo insertNodeAlgo) {
        this.insertNodeAlgo = insertNodeAlgo;
    }
    
    RAsteriskOverflowTreatmentAlgo(RAsteriskInsertNodeAlgo insertNodeAlgo) {
        this.insertNodeAlgo = insertNodeAlgo;
        
        getReInsertNodeAlgo().setInsertNodeAlgo(insertNodeAlgo);
    }
    
    Builder builder;
    public Builder getBuilder() {
        if(builder == null) {
            builder = new Builder();
        }
        return builder;
    }

    void execute(RTreeBase rTree) {
        if(rTree.getDebugMode() >= 1) {
            System.out.println("Overflow Treatment Algo Input");
            System.out.println("Node:");
            rTree.deepPrint(input.node);
            System.out.println("ReInsertCount:");
            System.out.println(input.reInsertCount);
            System.out.println("Level:");
            System.out.println(input.level);
            System.out.println();
        }

        if(!input.node.isRoot() && !visitedLevels.contains(input.level)) {
            visitedLevels.add(input.level);
            
            reInsertNodeAlgo.getBuilder().
                                setNode(input.node).
                                setReInsertCount(input.reInsertCount).
                                    build().
                                        execute(rTree);
        }
        else {
            if(input.node.isRoot()){
                rTree.createNewRootNode();
            }

            splitNodeAlgo.getBuilder().
                            setNode(input.node).
                                build().
                                    execute(rTree);

            Node parent = input.node.getParent();
                        
            splitNodeAlgo.getOutput().getNode1().pocessEntries();
            parent.getBuilder().addEntry(splitNodeAlgo.getOutput().getNode1());
            
            splitNodeAlgo.getOutput().getNode2().pocessEntries();
            parent.getBuilder().addEntry(splitNodeAlgo.getOutput().getNode2());
            
            parent.getBuilder().removeEntry(input.node);

            if(parent.ExceedMaxEntries()) {
                new RAsteriskOverflowTreatmentAlgo(insertNodeAlgo).getBuilder().
                                                        setNode(parent).
                                                        setLevel(input.level).
                                                        setReInsertCount(input.reInsertCount).
                                                            build().
                                                                execute(rTree);
            }
        }

        if(rTree.getDebugMode() >= 1) {
            System.out.println("Overflow Treatment Algo Output");
            System.out.println("Node:");
            rTree.deepPrint(output.node);
            System.out.println();
        }
    }
}
