/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab;

import java.util.ArrayList;
import java.util.List;

/**
 *
 * @author Lawal
 */
class RPlusPartitionAlgo {
    class Input {
        Node node;
         Node getNode() {
            return node;
        }
         void setNode(Node node) {
            this.node = node;
        }

        int fillFactor;
        int getFillFactor() {
            return fillFactor;
        }
        void setFillFactor(int fillFactor) {
            this.fillFactor = fillFactor;
        }
    }
    
    static class Output {
        boolean partitioned;
        boolean isPartitioned() {
            return partitioned;
        }
        void setPartitioned(boolean partitioned) {
            this.partitioned = partitioned;
        }
        
        Node node1;
         void setNode1(Node node1) {
            this.node1 = node1;
        }
         Node getNode1() {
            return node1;
        }

        Node node2;
         Node getNode2() {
            return node2;
        }
         void setNode2(Node node2) {
            this.node2 = node2;
        }

        int dim;
        int getDim() {
            return dim;
        }
        void setDim(int dim) {
            this.dim = dim;
        }

        double dimCut;
        double getDimCut() {
            return dimCut;
        }
        void setDimCut(double dimCut) {
            this.dimCut = dimCut;
        }

        double cost;
        double getCost() {
            return cost;
        }
        void setCost(double cost) {
            this.cost = cost;
        }
        
        List<Node> onBorderEntries = new ArrayList();
        List<Node> getOnBorderEntries() {
            return onBorderEntries;
        }
        void setOnBorderEntries(List<Node> onBorderEntries) {
            this.onBorderEntries = onBorderEntries;
        }
    }
    
    class Builder {
        Builder setNode(Node node) {
            RPlusPartitionAlgo.this.input.node = node;
            return this;
        }

        Builder setFillFactor(int fillFactor) {
            RPlusPartitionAlgo.this.input.fillFactor = fillFactor;
            return this;
        }

        RPlusPartitionAlgo build() {
            return RPlusPartitionAlgo.this;
        }
    }

    Input input = new Input();
    Input getInput() {
        return input;
    }
    void setInput(Input input) {
        this.input = input;
    }
    
    Output output = new Output();
    Output getOutput() {
        return output;
    }
    void setOutput(Output output) {
        this.output = output;
    }

    RPlusSweepAlgo sweepAlgo = new RPlusSweepAlgo();
    RPlusSweepAlgo getSweepAlgo() {
        return sweepAlgo;
    }
    void setSweepAlgo(RPlusSweepAlgo sweepAlgo) {
        this.sweepAlgo = sweepAlgo;
    }

    Builder builder;
    public Builder getBuilder() {
        if(builder == null) {
            builder = new Builder();
        }
        return builder;
    }

    void execute(RTreeBase rTree) {
        if(rTree.getDebugMode() >= 1) {
            System.out.println("RPlus Partition Algo Input");
            System.out.println("Node:");
            rTree.deepPrint(input.node);
            System.out.println("FillFactor:");
            System.out.println(Integer.toString(input.fillFactor));
            System.out.println();
        }

        if(input.node.getEntriesCount() < input.fillFactor)
        {
            output.partitioned = false;
        }
        else
        {
            double bestCost = Double.POSITIVE_INFINITY;
            for(int i = 0; i < input.node.getDim(); i++) {
                sweepAlgo.getBuilder().
                            setNode(input.node).
                            setDim(i).
                            setFillFactor(input.fillFactor).
                                build().
                                    execute(rTree);
                
                if(bestCost > sweepAlgo.getOutput().getCost()) {
                    bestCost = sweepAlgo.getOutput().getCost();
                    
                    output.partitioned = true;
                    
                    output.node1 = sweepAlgo.getOutput().getNode1();
                    output.node2 = sweepAlgo.getOutput().getNode2();
                    
                    output.dim = i;
                    output.dimCut = sweepAlgo.getOutput().getDimCut();
                    output.cost = sweepAlgo.getOutput().getCost();
                    
                    output.onBorderEntries = sweepAlgo.getOutput().getOnBorderEntries();
                }
            }
        }

        if(rTree.getDebugMode() >= 1) {
            System.out.println("RPlus Partition Algo Output");
            System.out.println("Node1:");
            rTree.deepPrint(output.node1);
            System.out.println("Node2:");
            rTree.deepPrint(output.node2);
            System.out.println("Dim:");
            System.out.println(Integer.toString(output.dim));
            System.out.println("DimCut:");
            System.out.println(Double.toString(output.dimCut));
            System.out.println("cost:");
            System.out.println(Double.toString(output.cost));
            
            System.out.println("On Border Entries:");
            for(Node entry:output.onBorderEntries) {
                rTree.deepPrint(entry);
            }
            
            System.out.println();
        }
    }
}
