/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab;

import java.util.ArrayList;
import java.util.List;

/**
 *
 * @author Lawal
 */
class RPlusSweepAlgo {
    enum SortEntriesBy {
        NEAREST_NEIGHBOR,
        MINIMAL_TOTAL_DISPLACMENT,
        MINIMAL_COVERAGE,
        MINIMAL_SPLIT
    }

    class Input {
        Node node;
         Node getNode() {
            return node;
        }
         void setNode(Node node) {
            this.node = node;
        }
         
        int dim;
        int getDim() {
            return dim;
        }
        void setDim(int dim) {
            this.dim = dim;
        }

        int fillFactor;
        int getFillFactor() {
            return fillFactor;
        }
        void setFillFactor(int fillFactor) {
            this.fillFactor = fillFactor;
        }        
    }
    
    class Output {
        Node node1;
         void setNode1(Node node1) {
            this.node1 = node1;
        }
        Node getNode1() {
            return node1;
        }

        Node node2;
         Node getNode2() {
            return node2;
        }
         void setNode2(Node node2) {
            this.node2 = node2;
        }

        double dimCut;
        double getDimCut() {
            return dimCut;
        }
        void setDimCut(double dimCut) {
            this.dimCut = dimCut;
        }
        
        double cost;
        double getCost() {
            return cost;
        }
        void setCost(double cost) {
            this.cost = cost;
        }
        
        List<Node> onBorderEntries;
        List<Node> getOnBorderEntries() {
            return onBorderEntries;
        }
        void setOnBorderEntries(List<Node> onBorderEntries) {
            this.onBorderEntries = onBorderEntries;
        }
    }
    
    class Builder {
        Builder setNode(Node node) {
            RPlusSweepAlgo.this.input.node = node;
            return this;
        }
         
        Builder setDim(int dim) {
            RPlusSweepAlgo.this.input.dim = dim;
            return this;
        }

        Builder setFillFactor(int fillFactor) {
            RPlusSweepAlgo.this.input.fillFactor = fillFactor;
            return this;
        }
        
        RPlusSweepAlgo build() {
            return RPlusSweepAlgo.this;
        }
    }

    Input input = new Input();
    Input getInput() {
        return input;
    }
    void setInput(Input input) {
        this.input = input;
    }
    
    Output output = new Output();
    Output getOutput() {
        return output;
    }
    void setOutput(Output output) {
        this.output = output;
    }

    Builder builder;
    public Builder getBuilder() {
        if(builder == null) {
            builder = new Builder();
        }
        return builder;
    }

    void execute(RTreeBase rTree) {
        if(rTree.getDebugMode() >= 1) {
            System.out.println("RPlus Sweep Algo Input");
            System.out.println("Node:");
            rTree.deepPrint(input.node);
            System.out.println("Dim:");
            System.out.println(Integer.toString(input.dim));
            System.out.println("FillFactor:");
            System.out.println(Integer.toString(input.fillFactor));
            System.out.println();
        }

        output.onBorderEntries = new ArrayList();
        
        Node node = rTree.createIntermediateNode();
        
        node.getBuilder().addEntriesTemporary(input.node.getEntries());
        
        node.getEntries().sort(new UtilitySortEntries(input.dim, Interval.Limit.UPPER, UtilitySortEntries.SortEntriesBy.LIMIT));
        
        double maxDimUpper = Double.NEGATIVE_INFINITY;

        output.node1 = rTree.createIntermediateNode();
        
        for(Node entry: node.getEntries().subList(0, input.fillFactor)) {
            output.node1.getBuilder().addEntryTemporary(entry);
            if(maxDimUpper < entry.getIntervalLimit(input.dim, Interval.Limit.UPPER)) {
                maxDimUpper = entry.getIntervalLimit(input.dim, Interval.Limit.UPPER);
            }
        }
        
        output.dimCut = maxDimUpper;
        
        output.node2 = rTree.createIntermediateNode();

        output.cost = 0;
        for(Node entry: node.getEntries().subList(input.fillFactor, node.getEntries().size())) {
            output.node2.getBuilder().addEntryTemporary(entry);
            if(entry.isDimCutBy(input.dim, output.dimCut)) {
                output.cost = output.cost + 1;
                output.onBorderEntries.add(entry);
            }
        }
        
        for(int i = 0; i < input.node.getDim(); i++) {
            output.node1.setIntervalA(i, input.node.getIntervalA(i));
            output.node1.setIntervalB(i, input.node.getIntervalB(i));
        }

        output.node1.setIntervalB(input.dim, output.dimCut);

        for(int i = 0; i < input.node.getDim(); i++) {
            output.node2.setIntervalA(i, input.node.getIntervalA(i));
            output.node2.setIntervalB(i, input.node.getIntervalB(i));
        }

        output.node2.setIntervalA(input.dim, output.dimCut);
        
        if(rTree.getDebugMode() >= 1) {
            System.out.println("RPlus Sweep Algo Output");
            System.out.println("Node1:");
            rTree.deepPrint(output.node1);
            System.out.println("Node2:");
            rTree.deepPrint(output.node2);
            System.out.println("DimCut:");
            System.out.println(Double.toString(output.dimCut));
            System.out.println("cost:");
            System.out.println(Double.toString(output.cost));
            
            System.out.println("On Border Entries:");
            for(Node entry:output.onBorderEntries) {
                rTree.deepPrint(entry);
            }
            
            System.out.println();
        }
    }
}
