/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab;

import java.io.IOException;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import spatial.tree.lab.utility.ByteBufferUtility;
import spatial.tree.lab.utility.FileUtility;

/**
 *
 * @author Lawal 
 */
public abstract class RTreeBase {
    Node rootNode;
    public Node getRootNode() {
        return rootNode;
    }
    public void setRootNode(Node rootNode) {
        this.rootNode = rootNode;
    }
    
    abstract void createNewRootNode();

    abstract Node createIntermediateNode();
    
    abstract public Node createDataNode();
    
    abstract void InsertData(Node data);
    
    int sequence = 0;
    public int getNewSequence() {
        return sequence++;
    }
    
    public int getNewSequence(int number) {
        sequence += number;
        return sequence - number;
    }

    public int getLeafLevel() {
        return rootNode.getLastLevel();
    }
    
    public int getLeafNodeLevel() {
        return rootNode.getLastLevel() == 0 ? 0 : rootNode.getLastLevel() - 1;
    }

    int debugMode = 0;
    public int getDebugMode() {
        return debugMode;
    }
    public void setDebugMode(int debugMode) {
        this.debugMode = debugMode;
    }
     
    public int getMaxNodeMemorySize() {
        return rootNode.getMaxNodeMemorySize();
    }

    public long getMemorySize() {
        return rootNode.deepGetMemorySize();
    }

    public long getHardDriveSize() {
        return rootNode.deepGetHardDriveSize();
    }

    public void deepPrint(Node node) {
        if(node == null) {
            System.out.println("null");   
        } 
        else {
            System.out.println(node.toFullString(node.getLevel()));
            for(Node child: node.getEntries()) {
                deepPrint(child, node.getLevel() + 1);
            }
        }
    }

    public void deepPrint(Node node, int level) {
        System.out.println(node.toFullString(level));
        for(Node child: node.getEntries()) {
            deepPrint(child, level + 1);
        }
    }

    public void print() {
        System.out.println("*********** Tree ************");
        deepPrint(rootNode);
    }
    
    public void deepPrintSelected(Node node) {
        if(node == null) {
            System.out.println("null");   
        } 
        else {
            System.out.println(node.toFullString(node.getLevel()));
            for(Node child: node.getEntries()) {
                if(child.getIsSelected()) {
                    deepPrintSelected(child, node.getLevel() + 1);
                }
            }
        }
    }

    public void deepPrintSelected(Node node, int level) {
        System.out.println(node.toFullString(level));
        for(Node child: node.getEntries()) {
            if(child.getIsSelected()) {
                deepPrintSelected(child, level + 1);
            }
        }
    }

    public void printSelected() {
        System.out.println("*********** Tree ************");
        deepPrintSelected(rootNode);
    }

    public long getNodeCount() {
        return rootNode.deepGetNodesCount();
    }
    
    public int getLeafCount() {
        return rootNode.deepGetLeavesCount();
    }

    public RQueryResult getNodes(MinBoundingRectangle minBoundingRectangle) {
        long startTime = System.nanoTime();
        RQueryResult results = rootNode.getNodes(this, minBoundingRectangle);
        long endTime = System.nanoTime();

        long duration = (endTime - startTime);         
        results.setExecutionTime(duration);
        return results;
     }

    public RQueryResult getNodes(int d, Interval interval) {
        long startTime = System.nanoTime();
        RQueryResult results = rootNode.getNodes(this, d, interval);
        long endTime = System.nanoTime();

        long duration = (endTime - startTime);         
        results.setExecutionTime(duration);
        return results;
     }

    RandomAccessFile file;
    public RandomAccessFile getFile() {
        return file;
    }
    void setFile(RandomAccessFile file) throws IOException {
        if(file != null) {
            this.file = file;
            sequence = (int)(file.length() / (long)ByteBufferUtility.MEMORY_PAGE_SIZE);
        }
    }
    
    public void prepareToSave() {
        rootNode.deepAllocateSequence(this);
    }
    
    public void writeToByteArray(RandomAccessFile out) throws IOException {
        rootNode.deepWriteToByteArray(out); 
    }
     
     void assertTree() {
    }
}
