/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab;

import java.util.ArrayList;
import java.util.List;

/**
 *
 * @author Lawal
 */
class XTreeOverflowTreatmentAlgo {
    class Input {
        Node node;
        Node getNode() {
            return node;
        }
        void setNode(Node node) {
            this.node = node;
        }
        
        int level;
        int getLevel() {
            return level;
        }
        void setLevel(int level) {
            this.level = level;
        }

        int reInsertCount;
        int getReInsertCount() {
            return reInsertCount;
        }
        void setReInsertCount(int reInsertCount) {
            this.reInsertCount = reInsertCount;
        }
        
        double maxOverlap;
        double getMaxOverlap() {
            return maxOverlap;
        }
        void setMaxOverlap(double maxOverlap) {
            this.maxOverlap = maxOverlap;
        }

        int fanOut;
        int getFanOut() {
            return fanOut;
        }
        void setFanOut(int fanOut) {
            this.fanOut = fanOut;
        }
    }

    static class Output {
        Node node;
        Node getNode() {
            return node;
        }
        void setNode(Node node) {
            this.node = node;
        }
    }

    class Builder {
        Builder setNode(Node node) {
            XTreeOverflowTreatmentAlgo.this.input.node = node;
            return this;
        }

        Builder setLevel(int level) {
            XTreeOverflowTreatmentAlgo.this.input.level = level;
            return this;
        }

        Builder setReInsertCount(int reInsertCount) {
            XTreeOverflowTreatmentAlgo.this.input.reInsertCount = reInsertCount;
            return this;
        }
        
        Builder setMaxOverlap(double maxOverlap) {
            XTreeOverflowTreatmentAlgo.this.input.maxOverlap = maxOverlap;
            return this;
        }

        Builder setFanout(int fanout) {
            XTreeOverflowTreatmentAlgo.this.input.fanOut = fanout;
            return this;
        }

        XTreeOverflowTreatmentAlgo build() {
            return XTreeOverflowTreatmentAlgo.this;
        }
    }
    
    Input input = new Input();
    Input getInput() {
        return input;
    }
    void setInput(Input input) {
        this.input = input;
    }
    
    Output output = new Output();
    Output getOutput() {
        return output;
    }
    void setOutput(Output output) {
        this.output = output;
    }
    
    List<Integer> visitedLevels = new ArrayList<>();
    List<Integer> getVisitedLevels() {
        return visitedLevels;
    }
    void setVisitedLevels(List<Integer> visitedLevels) {
        this.visitedLevels = visitedLevels;
    }

    XTreeSplitNodeAlgo splitNodeAlgo = new XTreeSplitNodeAlgo();
    XTreeSplitNodeAlgo getSplitNodeAlgo() {
        return splitNodeAlgo;
    }
    void setSplitNodeAlgo(XTreeSplitNodeAlgo splitNodeAlgo) {
        this.splitNodeAlgo = splitNodeAlgo;
    }

    XTreeReInsertNodeAlgo reInsertNodeAlgo = new XTreeReInsertNodeAlgo();
    XTreeReInsertNodeAlgo getReInsertNodeAlgo() {
        return reInsertNodeAlgo;
    }
    void setReInsertNodeAlgo(XTreeReInsertNodeAlgo reInsertNodeAlgo) {
        this.reInsertNodeAlgo = reInsertNodeAlgo;
    }
    
    XTreeInsertNodeAlgo insertNodeAlgo;
    XTreeInsertNodeAlgo getInsertNodeAlgo() {
        return insertNodeAlgo;
    }
    void setInsertNodeAlgo(XTreeInsertNodeAlgo insertNodeAlgo) {
        this.insertNodeAlgo = insertNodeAlgo;
    }
    
    XTreeOverflowTreatmentAlgo(XTreeInsertNodeAlgo insertNodeAlgo) {
        this.insertNodeAlgo = insertNodeAlgo;
        
        getReInsertNodeAlgo().setInsertNodeAlgo(insertNodeAlgo);
    }

    Builder builder;
    public Builder getBuilder() {
        if(builder == null) {
            builder = new Builder();
        }
        return builder;
    }

    void execute(RTreeBase rTree) {
        if(rTree.getDebugMode() >= 1) {
            System.out.println("Insert Node Algo Input");
            System.out.println("Node:");
            rTree.deepPrint(input.node);
            System.out.println("Level:");
            System.out.println(input.level);
            System.out.println("ReInsertCount:");
            System.out.println(input.reInsertCount);
            System.out.println("MaxOverlap:");
            System.out.println(input.maxOverlap);
            System.out.println("FanOut:");
            System.out.println(input.fanOut);
            System.out.println();
        }

        if(!input.node.isRoot() && !visitedLevels.contains(input.level)) {
            visitedLevels.add(input.level);
            
            reInsertNodeAlgo.getBuilder().
                    setNode(input.node).
                    setReInsertCount(input.reInsertCount).
                        build().
                            execute(rTree);
        }
        else {
            splitNodeAlgo.getBuilder().
            setNode(input.node).
            setReInsertCount(input.reInsertCount).
            setMaxOverlap(input.maxOverlap).
            setFanout(input.fanOut).
                build().
                    execute(rTree);

            if(splitNodeAlgo.getOutput().isSucceded()) {
                if(input.node.isRoot()) {
                    rTree.createNewRootNode();
                }
                
                Node parent = input.node.getParent();

                splitNodeAlgo.getOutput().getNode1().pocessEntries();
                parent.getBuilder().addEntry(splitNodeAlgo.getOutput().getNode1());
                
                splitNodeAlgo.getOutput().getNode2().pocessEntries();
                parent.getBuilder().addEntry(splitNodeAlgo.getOutput().getNode2());

                parent.getBuilder().removeEntry(input.node);

                if(parent.ExceedMaxEntries()) {
                    new XTreeOverflowTreatmentAlgo(insertNodeAlgo).getBuilder().
                                                        setNode(parent).
                                                        setLevel(input.level).
                                                        setReInsertCount(input.reInsertCount).
                                                        setMaxOverlap(input.maxOverlap).
                                                        setFanout(input.fanOut).
                                                            build().
                                                                execute(rTree);
                }
            }
            else {
                input.node.getBuilder().setIsSuperNode();
            }
        }

        if(rTree.getDebugMode() >= 1) {
            System.out.println("Insert Node Algo Output");
            System.out.println("Node:");
            rTree.deepPrint(output.node);
        }
    }
}
