/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab.experiment;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Scanner;
import spatial.tree.lab.Node;
import spatial.tree.lab.RPlusTree;
import spatial.tree.lab.RTree;
import spatial.tree.lab.XTree;
import spatial.tree.lab.adapters.TextFileAdapter;
import spatial.tree.lab.utility.ByteBufferUtility;
import spatial.tree.lab.utility.FileUtility;

/**
 *
 * @author Mahmoud
 */
public class RPlusTreeCreation {
    public static void main(String[] args) throws IOException, InterruptedException {
        System.out.println("********** RPlus Tree Creation **************\n");
        
        Scanner sc = new Scanner(System.in);
        System.out.println("Enter number of dimensions: ");
        int d = sc.nextInt();
        
        System.out.println("Enter M: ");
        int M = sc.nextInt();
        
        System.out.println("Enter m: ");
        int m = sc.nextInt();

        System.out.println("Enter fill factor: ");
        int fillFactor = sc.nextInt();

        System.out.println("Enter file name: ");
        String f = sc.next();

        String treeFileName = f.replace(".dataset", "") + "_" +
                                Integer.toString(M) + "_" +
                                Integer.toString(m) +
                                ".rblustree";
        
        System.out.println("Start from: ");
        long startFrom = sc.nextLong();
        
        System.out.println("Number of records: ");
        long numberOfRecords = sc.nextLong();

        DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy/MM/dd HH:mm:ss");  
        System.out.println("Start @ Time:" + dtf.format(LocalDateTime.now())); 

        FileUtility fileUtility = new FileUtility();
        RandomAccessFile treeFile = fileUtility.openRandomAccessFile(treeFileName);
        
        RPlusTree rtree = new RPlusTree(treeFile, d, m, M, fillFactor);
        rtree.setDebugMode(0);
        
        if(startFrom > 0) {
            rtree.getRootNode().setSequence(0);
            rtree.getRootNode().setIsLoaded(false);
            rtree.getRootNode().loadNode();
        }
                
        long counter = 0;

        TextFileAdapter textFileAdapter = new TextFileAdapter();
        textFileAdapter.setFileName(f);
        textFileAdapter.start(rtree);
        
        long numberOfRecordsAdded = 0;

        Node leaf = textFileAdapter.getNext(rtree);
        while(leaf != null) {
            counter++;
            
            if(counter > startFrom) {
                rtree.InsertData(leaf);
                leaf = textFileAdapter.getNext(rtree);
                
                numberOfRecordsAdded++;
            }
            
            if(counter == startFrom + numberOfRecords ) {
                System.out.println("********** Save the Tree **************\n");
                fileUtility.saveToFile(rtree);
                treeFile.close();
                System.out.println("Finish @ Time:" + dtf.format(LocalDateTime.now())); 
                break;
            }
            else {
                if(numberOfRecordsAdded % 100000 == 0) {
                    System.out.println("Parital Save the " + numberOfRecordsAdded + " records.");

                    fileUtility.saveToFile(rtree);
                    treeFile.close();
                    
                    System.out.println("@ Time:" + dtf.format(LocalDateTime.now())); 

                    treeFile = fileUtility.openRandomAccessFile(treeFileName);
                    rtree = new RPlusTree(treeFile, d, m, M, fillFactor);
                    rtree.getRootNode().setSequence(0);
                    rtree.getRootNode().setIsLoaded(false);
                    rtree.getRootNode().loadNode();
                }
            }
        }
        
        textFileAdapter.end();
        
        System.out.println("********** Tree Output **************\n");
        
        System.out.println("Number of Levels:" + Integer.toString(rtree.getLeafLevel()));
        System.out.println("Leaves Count:" + Long.toString(counter));
        System.out.println("Nonleaf Nodes Count:" + Long.toString(rtree.getNodeCount()));
        System.out.println("Single Node Memory Size:" + Integer.toString(rtree.getMaxNodeMemorySize()));
        System.out.println("Memory Page Size:" + Integer.toString(ByteBufferUtility.MEMORY_PAGE_SIZE));
        System.out.println("Memory Size:" + Long.toString(rtree.getMemorySize()));
        System.out.println("Hard Drive Size:" + Long.toString(rtree.getHardDriveSize()));
        System.out.println();
        
        treeFile.close();
    }    
}
