/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab.experiment;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import spatial.tree.lab.MinBoundingRectangle;
import spatial.tree.lab.Node;
import spatial.tree.lab.RAsteriskRTree;
import spatial.tree.lab.RQueryResult;
import spatial.tree.lab.RTree;
import spatial.tree.lab.adapters.TextFileAdapter;
import spatial.tree.lab.utility.ByteBufferUtility;
import spatial.tree.lab.utility.FileUtility;

/**
 *
 * @author Mahmoud
 */
public class RandomQueryExecution {
    public static void main(String[] args) throws IOException, InterruptedException {
        System.out.println("********** Query Results **************\n");
        
        Scanner sc = new Scanner(System.in);
        System.out.println("Enter number of dimensions: ");
        int d = sc.nextInt();
        
        System.out.println("Enter M: ");
        int M = sc.nextInt();
        
        System.out.println("Enter m: ");
        int m = sc.nextInt();

        System.out.println("Enter tree file name: ");
        String treeFileName = sc.next();
        
        System.out.println("Enter query file name: ");
        String queryFileName = sc.next();
        
        FileUtility fileUtility = new FileUtility();
        RandomAccessFile treeFile = fileUtility.openRandomAccessFile(treeFileName);
        
        RTree rTree = new RTree(treeFile, d, m, M);
        rTree.setDebugMode(0);
        rTree.getRootNode().setSequence(0);
        rTree.getRootNode().setIsLoaded(false);
        rTree.getRootNode().loadNode();
                
        TextFileAdapter textFileAdapter = new TextFileAdapter();
        textFileAdapter.setFileName(queryFileName);
        textFileAdapter.start(rTree);

        List<MinBoundingRectangle> queries = new ArrayList<>();
        Node mrp = textFileAdapter.getNext(rTree);
        while(mrp != null) {
            queries.add(mrp.getMinBoundingRectangle());
            mrp = textFileAdapter.getNext(rTree);
        }
        
        textFileAdapter.end();
        treeFile.close();
        
        System.out.println("********** Execute Random Query **************\n");
        double avgPages = 0;
        double avgNodeVisited = 0;
        double avgExecutionTime = 0;
        for(int i = 0; i < queries.size(); i++) {
            System.out.println("Query " + Integer.toString(i + 1) + ":");
            
            treeFile = fileUtility.openRandomAccessFile(treeFileName);
            rTree = new RTree(treeFile, d, m, M);
            rTree.setDebugMode(0);
            rTree.getRootNode().setSequence(0);
            rTree.getRootNode().setIsLoaded(false);
            rTree.getRootNode().loadNode();            
            
            System.out.println("MBR:" + queries.get(i).toFullString());
            
            RQueryResult rQueryResult = rTree.getNodes(queries.get(i));
            rQueryResult.markNodesAsSelected();
            
            int pages = 0;
            for(Node node:rQueryResult.getVisitedNodes()) {
                pages += node.getPagesCount();
            }
            
            System.out.println("Number of Pages Visited:" + Integer.toString(pages));
            System.out.println("Number of Nodes Visited:" + Integer.toString(rQueryResult.getVisitedNodes().size()));
            System.out.println("Number of Data Leafs:" + Integer.toString(rQueryResult.getLeaves().size()));
            System.out.println("Execution Time in Nano Second:" + Long.toString(rQueryResult.getExecutionTime()));
            
            //rTree.printSelected();
            
            avgPages += pages / queries.size();
            avgNodeVisited += rQueryResult.getVisitedNodes().size() / queries.size();
            avgExecutionTime += rQueryResult.getExecutionTime() / queries.size();
            
            rQueryResult.unmarkNodesAsSelected();
            
            treeFile.close();
        }
        
        System.out.println();
        System.out.println("********** Execute Random Query Summary **************\n");
        System.out.println("Avarage Number of Pages Visited:" + Double.toString(avgPages));
        System.out.println("Avarage Number of Nodes Visited:" + Double.toString(avgNodeVisited));
        System.out.println("Avarage Execution Time in Nano Second:" + Double.toString(avgExecutionTime));
    }    
}
