/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab.query;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import spatial.tree.lab.MinBoundingRectangle;
import spatial.tree.lab.Node;
import spatial.tree.lab.RQueryResult;
import spatial.tree.lab.RTreeBase;
import spatial.tree.lab.XTree;
import spatial.tree.lab.utility.DefuzzificationMethod;
import spatial.tree.lab.utility.FuzzyUtility;

/**
 *
 * @author Mahmoud
 */
public class SkylineQuery {
    private Object rootNode;
    public RQueryResult getNodes(RTreeBase xTree, MinBoundingRectangle minBoundingRectangle, double stepThreshold, double neighbourRange) {
        
        long startTime = System.nanoTime();
        RQueryResult results = xTree.getNodes(minBoundingRectangle);
        
        return results;
    }
    
    public List<Node> getSkyLine(RTreeBase xTree, MinBoundingRectangle minBoundingRectangle, double stepLength, double neighbourMaxDistance, double maxEpselon, DefuzzificationMethod method) {
        List<Node> nodes = xTree.getNodes(minBoundingRectangle).getLeaves();
        
        FuzzyUtility fu = new FuzzyUtility();
        
        double epselon = Double.POSITIVE_INFINITY;
        
        while(epselon > maxEpselon) {
            epselon = 0;
            for(Node node1:nodes) {
                int impDim = node1.getImpreciseDim();

                List<Double> memberShipFunction = new ArrayList<>();
                List<Double> range = new ArrayList<>();

                for(double perDim = node1.getMinBoundingRectangle().getIntervalA(impDim); perDim < node1.getMinBoundingRectangle().getIntervalB(impDim); perDim += stepLength) {
                    range.add(perDim);
                    
                    MinBoundingRectangle mbr = node1.getMinBoundingRectangle().deepClone();
                    mbr.getInterval(impDim).setAB(perDim, perDim);

                    List<Node> neighbourNode = new ArrayList<>();
                    
                    for(Node node2:nodes) {
                        if(node1 != node2) {
                            if(mbr.getPreciseCenterDistance(node2.getMinBoundingRectangle()) < neighbourMaxDistance) {
                               neighbourNode.add(node2);
                            }
                        }
                    }

                    double sum = 0;
                    for(Node child:neighbourNode) {
                        sum += 1.0 / (1.0 + node1.getMinBoundingRectangle().getPreciseCenterDistance(child.getMinBoundingRectangle()));
                    }

                    if(neighbourNode.size() == 0) {
                        sum = 0;   
                    }
                    else {
                        sum = sum / neighbourNode.size();
                    }

                    memberShipFunction.add(sum);
                }
                
                double oldC = node1.getMinBoundingRectangle().getInterval(impDim).getC();
                double newC = fu.Defuzzify(stepLength, range, memberShipFunction, method);

                epselon += Math.pow(newC - oldC, 2); 
                
                node1.getMinBoundingRectangle().getInterval(impDim).setC(newC);
            }
        }
        
        List<Node> skyline = new ArrayList<>();
        
        for(Node node1:nodes) {
            int count = 0;
            for(Node node2:nodes) {
                if(node1 != node2) {
                    if(node1.getChildren().size() == 0 && node2.getChildren().size() == 0 &&
                       isPareto(node1, node2)) {
                        count++;
                    }

                    if(node1.getChildren().size() > 0 && node2.getChildren().size() == 0 &&
                       isPareto(node1.getChildren(), node2)) {
                        count++;
                    }

                    if(node1.getChildren().size() == 0 && node2.getChildren().size() > 0 &&
                       isPareto(node1, node2.getChildren())) {
                        count++;
                    }

                    if(node1.getChildren().size() > 0 && node2.getChildren().size() > 0 &&
                       isPareto(node1.getChildren(), node2.getChildren())) {
                        count++;
                    }
                }
            }
            
            if(count == nodes.size() - 1) {
                skyline.add(node1);
            }
        }
        
        return skyline;
    }
    
    public boolean isPareto(Node node1, Node node2) {
        for(int i = 0; i < node1.getDim(); i++) {
            if(node1.getIntervalC(i) < node2.getIntervalC(i)) {
                return true;
            }
        }
        
        return false;
    }

    public boolean isPareto(Node node1, List<Node> nodes2) {
        for(Node node2:nodes2) {
            for(int i = 0; i < node1.getDim(); i++) {
                if(node1.getIntervalC(i) < node2.getIntervalC(i)) {
                    return true;
                }
            }
        }
        
        return false;
    }

    public boolean isPareto(List<Node> nodes1, Node node2) {
        for(Node node1:nodes1) {
            for(int i = 0; i < node1.getDim(); i++) {
                if(node1.getIntervalC(i) < node2.getIntervalC(i)) {
                    return true;
                }
            }
        }
        
        return false;
    }

    public boolean isPareto(List<Node> nodes1, List<Node> nodes2) {
        for(Node node1:nodes1) {
            for(Node node2:nodes2) {
                for(int i = 0; i < node1.getDim(); i++) {
                    if(node1.getIntervalC(i) < node2.getIntervalC(i)) {
                        return true;
                    }
                }
            }
        }
        
        return false;
    }
        
    
    public static String getToPrint(double a, double b, double c) {
        if(a == b) return "" + a;
        return a + "-" + b + " (" + c + ")";
    }
    
    /**
     *
     * @param args
     */
    public static void main(String[] args) throws IOException {
        SkylineQuery skyline = new SkylineQuery();
        
        System.out.println("********** XTree **************\n");

        List<Node> input = new ArrayList<>();
        
        int dim = 2;
        int m = 2;
        int M = 4;
        int reInsertCount = 3;
        double maxOverlap = 0.2;
        int fanout =  3;
        XTree rtree = new XTree(null,dim, m, M, reInsertCount, maxOverlap, fanout);
        rtree.setDebugMode(0);

                Node a = rtree.createDataNode().getBuilder().
                                            setId("A").
                                            setInterval(0, 3, 5).
                                            setInterval(1, 6).
                                            build();
        rtree.InsertData(a);
        rtree.print();
        
        System.out.println("**********************************\n");
        
         
        Node b = rtree.createDataNode().getBuilder().
                                            setId("B").
                                            setInterval(0, 1).
                                            setInterval(1, 2, 7).
                                            build();
        rtree.InsertData(b);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node c = rtree.createDataNode().getBuilder().
                                            setId("C").
                                            setInterval(0, 7, 9).
                                            setInterval(1, 6).
                                                    build();
        rtree.InsertData(c);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node d = rtree.createDataNode().getBuilder().
                                            setId("D").
                                            setInterval(0, 5).
                                            setInterval(1, 5, 10).
                                                    build();
        rtree.InsertData(d);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node e = rtree.createDataNode().getBuilder().
                                            setId("E").
                                            setInterval(0, 1, 6).
                                            setInterval(1, 5).
                                            build();
        rtree.InsertData(e);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node f = rtree.createDataNode().getBuilder().
                                            setId("F").
                                            setInterval(0, 4).
                                            setInterval(1, 3, 5).
                                            build();
        rtree.InsertData(f);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node g = rtree.createDataNode().getBuilder().
                                            setId("G").
                                            setInterval(0, 4).
                                            setInterval(1, 7, 12).
                                            build();
        rtree.InsertData(g);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node h = rtree.createDataNode().getBuilder().
                                            setId("H").
                                            setInterval(0, 6, 9).
                                            setInterval(1, 3).
                                                    build();
        rtree.InsertData(h);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node i = rtree.createDataNode().getBuilder().
                                            setId("I").
                                            setInterval(0, 1, 10).
                                            setInterval(1, 8).
                                                    build();
        rtree.InsertData(i);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node j = rtree.createDataNode().getBuilder().
                                            setId("J").
                                            setInterval(0, 11).
                                            setInterval(1, 4, 6).
                                            build();
        rtree.InsertData(j);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node k = rtree.createDataNode().getBuilder().
                                            setId("K").
                                            setInterval(0, 7).
                                            setInterval(1, 1, 4).
                                                    build();
        rtree.InsertData(k);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node l = rtree.createDataNode().getBuilder().
                                            setId("L").
                                            setInterval(0, 4, 9).
                                            setInterval(1, 5).
                                                    build();
        rtree.InsertData(l);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node ml = rtree.createDataNode().getBuilder().
                                            setId("M").
                                            setInterval(0, 10).
                                            setInterval(1, 3, 8).
                                            build();
        rtree.InsertData(ml);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node n = rtree.createDataNode().getBuilder().
                                            setId("N").
                                            setInterval(0, 4, 6).
                                            setInterval(1, 5).
                                            build();
        rtree.InsertData(n);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node o = rtree.createDataNode().getBuilder().
                                            setId("O").
                                            setInterval(0, 8).
                                            setInterval(1, 7, 12).
                                                    build();
        rtree.InsertData(o);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node p = rtree.createDataNode().getBuilder().
                                            setId("P").
                                            setInterval(0, 9).
                                            setInterval(1, 5, 11).
                                            build();
        rtree.InsertData(p);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node q = rtree.createDataNode().getBuilder().
                                            setId("Q").
                                            setInterval(0, 7, 12).
                                            setInterval(1, 3).
                                            build();
        rtree.InsertData(q);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node r = rtree.createDataNode().getBuilder().
                                            setId("R").
                                            setInterval(0, 6).
                                            setInterval(1, 2, 5).
                                            build();
        rtree.InsertData(r);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node s = rtree.createDataNode().getBuilder().
                                            setId("S").
                                            setInterval(0, 10, 12).
                                            setInterval(1, 7).
                                            build();
        rtree.InsertData(s);
        rtree.print();
        
        System.out.println("**********************************\n");
        
                
        Node t = rtree.createDataNode().getBuilder().
                                            setId("T").
                                            setInterval(0, 8).
                                            setInterval(1, 4, 6).
                                                    build();
        rtree.InsertData(t);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node u = rtree.createDataNode().getBuilder().
                                            setId("U").
                                            setInterval(0, 10, 12).
                                            setInterval(1, 3).
                                            build();
        rtree.InsertData(u);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node v = rtree.createDataNode().getBuilder().
                                            setId("V").
                                            setInterval(0, 11).
                                            setInterval(1, 3, 5).
                                            build();
        rtree.InsertData(v);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node w = rtree.createDataNode().getBuilder().
                                            setId("W").
                                            setInterval(0, 3, 9).
                                            setInterval(1, 5).
                                                    build();
        rtree.InsertData(w);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node x = rtree.createDataNode().getBuilder().
                                            setId("X").
                                            setInterval(0, 1, 8).
                                            setInterval(1, 6).
                                            build();
        rtree.InsertData(x);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node y = rtree.createDataNode().getBuilder().
                                            setId("y").
                                            setInterval(0, 7, 9).
                                            setInterval(1, 4).
                                                    build();
        rtree.InsertData(y);
        rtree.print();
        
        System.out.println("**********************************\n");
        
        Node z = rtree.createDataNode().getBuilder().
                                            setId("Z").
                                            setInterval(0, 9, 12).
                                            setInterval(1, 2).
                                            build();
        rtree.InsertData(z);
                
        rtree.print();
        
        
        System.out.println("********** SkyLine MBR **************");
        
        MinBoundingRectangle rootMbr = new MinBoundingRectangle(2);
        rootMbr.getInterval(0).setA(0);
        rootMbr.getInterval(0).setB(12);
        rootMbr.getInterval(1).setA(0);
        rootMbr.getInterval(1).setB(12);
        
        List<Node> skylineResult = skyline.getSkyLine(rtree, rootMbr, .1, 5, .01, DefuzzificationMethod.SOM);

        System.out.println("Data");
        for(Node node:rtree.getNodes(rootMbr).getLeaves()) {
            System.out.println(node.getId() + ": " + getToPrint (node.getIntervalA(0), node.getIntervalB(0) , node.getIntervalC(0)) + ", " + getToPrint (node.getIntervalA(1), node.getIntervalB(1) , node.getIntervalC(1)));
        }

        System.out.println("Skyline");
        for(Node node:skylineResult) {
            System.out.println(node.getId() + ": " + getToPrint (node.getIntervalA(0), node.getIntervalB(0) , node.getIntervalC(0)) + ", " + getToPrint (node.getIntervalA(1), node.getIntervalB(1) , node.getIntervalC(1)));
        }
    }
}




