/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab.testing;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.List;
import spatial.tree.lab.MinBoundingRectangle;
import spatial.tree.lab.Node;
import spatial.tree.lab.RQueryResult;
import spatial.tree.lab.RTree;
import spatial.tree.lab.XTree;
import spatial.tree.lab.adapters.TextFileAdapter;
import spatial.tree.lab.dataset.SnytheticDataSetGenerator;
import spatial.tree.lab.query.RandomUniformQueryGenerator;
import spatial.tree.lab.query.RandomUniformSelectivityQueryGenerator;
import spatial.tree.lab.utility.ByteBufferUtility;
import spatial.tree.lab.utility.FileUtility;
import spatial.tree.lab.utility.RTreeUtility;

/**
 *
 * @author AbbeyMalay
 */
public class ExperimentTemplate {
        public static void main(String[] args) throws IOException, InterruptedException{
        System.out.println("********** Indices over Range Query: selectivity=0.1; RGQ = 10; dimension=17 and datasize= from 100k to 1M **************\n");

        // RTree
        int dim = 17;
        int m = 6;
        int M = 11;
        
        RTree rtree = new RTree(null,dim, m, M);
        rtree.setDebugMode(0);
        
        // RTree*
        //int dim = 6;
        //int m = 2;
        //int M = 5;
        //int reInsertCount = 2;
        //RAsteriskRTree rtree = new RAsteriskRTree(dim, m, M, reInsertCount);
        //rtree.setDebugMode(0);
        
        // RTree X
        //int dim = 5;
        //int m = 2;
        //int M = 5;
        //int reInsertCount = 3;
        //double maxOverlap = .2;
        //int fanout =  4;
        //XTree rtree = new XTree(dim, m, M, reInsertCount, maxOverlap, fanout);
        //rtree.setDebugMode(0);
        
        // RTree +
        //int dim = 6;
        //int m = 2;
        //int M = 4;
        //int fillFactor = 3;
        //RPlusTree rtree = new RPlusTree(dim, m, M, fillFactor);
        //rtree.setDebugMode(0);
        
        //SnytheticDataSetGenerator generator = new SnytheticDataSetGenerator();
        //generator.createDataSet("d:\\Tree2000.txt", 5,  1000, 20.0, 'E');

        TextFileAdapter textFileAdapter = new TextFileAdapter();
        textFileAdapter.setFileName("f:\\100k_PA_inde.txt");
        
        textFileAdapter.start(rtree);
        
        Node leaf = textFileAdapter.getNext(rtree);
        while(leaf != null) {
            rtree.InsertData(leaf);
            leaf = textFileAdapter.getNext(rtree);
        }
        
        textFileAdapter.end();
        
        System.out.println("********** Print Tree **************\n");
        //rtree.print();
        
        System.out.println("********** Tree Report **************\n");
        
        System.out.println("Number of Levels:" + Integer.toString(rtree.getLeafLevel()));
        //System.out.println("Nonleaf Nodes Count:" + Integer.toString(rtree.getNodeCount()));
        System.out.println("Leaves Count:" + Integer.toString(rtree.getLeafCount()));
        System.out.println("Single Node Memory Size:" + Integer.toString(rtree.getMaxNodeMemorySize()));
        System.out.println("Memory Page Size:" + Integer.toString(ByteBufferUtility.MEMORY_PAGE_SIZE));
        System.out.println("Memory Size:" + Long.toString(rtree.getMemorySize()));
        System.out.println("Hard Drive Size:" + Long.toString(rtree.getHardDriveSize()));
        System.out.println();

                
        System.out.println("********** Save the Tree **************\n");
        FileUtility fileUtility = new FileUtility();
        //fileUtility.saveToFile(rtree, "d:\\100k_PA_inde_index.txt");
        
        System.out.println("********** Get Root MRB **************\n");
        RTreeUtility utlity = new RTreeUtility();
        Node root = utlity.getRootNode("d:\\100k_PA_inde_index.txt");
        System.out.println("Root: " + root.toFullString());
        

        System.out.println("********** Generate Random Query **************\n");        
        RandomUniformSelectivityQueryGenerator randSelectivityQuery = new RandomUniformSelectivityQueryGenerator();
        List<MinBoundingRectangle> randQueries = randSelectivityQuery.generateRandomQueries(rtree, 10, 1000);
 
        System.out.println("********** Execute Random Query **************\n");        
        
        // Avarage
        int avgPages = 0;
        int avgVisitedNodes = 0;
        long avgExecutionTime = 0;
        // Avarage
        
        for(int i = 0; i < randQueries.size(); i++) {
            System.out.println((i + 1) + "- " + randQueries.get(i).toFullString());
            
            RandomAccessFile file = fileUtility.openRandomAccessFile("d:\\100k_PA_inde_index.txt");
            RTree rtreeToQuery = new RTree(null, dim, m, M);
            MinBoundingRectangle rqueryMbr = randQueries.get(i);
            RQueryResult rQueryResult = rtreeToQuery.getNodes(rqueryMbr);
            rQueryResult.markNodesAsSelected();
            
            int pages = 0;
            for(Node node:rQueryResult.getVisitedNodes()) {
                pages += node.getPagesCount();
            }
            
            System.out.println("Number of Pages Visited:" + Integer.toString(pages));
            System.out.println("Number of Nodes Visited:" + Integer.toString(rQueryResult.getVisitedNodes().size()));
            System.out.println("Number of Data Leafs:" + Integer.toString(rQueryResult.getLeaves().size()));
            System.out.println("Execution Time in Nano Second:" + Long.toString(rQueryResult.getExecutionTime()));
            
            //rtreeToQuery.printSelected();
            
            rQueryResult.unmarkNodesAsSelected();
            
            file.close();
            
            // Avarage
            avgPages += pages;
            avgVisitedNodes += rQueryResult.getVisitedNodes().size();
            avgExecutionTime += rQueryResult.getExecutionTime();            
            // Avarage
        }
        
        // Avarage
        System.out.println("Avarage Number of Pages Visited:" + Integer.toString(avgPages));
        System.out.println("Avarage Number of Nodes Visited:" + Integer.toString(avgVisitedNodes));
        System.out.println("Avarage Execution Time in Nano Second:" + Long.toString(avgExecutionTime));
        // Avarage
    }

}
