/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab.testing;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.List;
import spatial.tree.lab.MinBoundingRectangle;
import spatial.tree.lab.Node;
import spatial.tree.lab.RAsteriskRTree;
import spatial.tree.lab.RPlusTree;
import spatial.tree.lab.RQueryResult;
import spatial.tree.lab.RTree;
import spatial.tree.lab.XTree;
import spatial.tree.lab.adapters.TextFileAdapter;
import spatial.tree.lab.dataset.SnytheticDataSetGenerator;
import spatial.tree.lab.query.RandomUniformQueryGenerator;
import spatial.tree.lab.query.RandomUniformSelectivityQueryGenerator;
import spatial.tree.lab.utility.ByteBufferUtility;
import spatial.tree.lab.utility.FileUtility;
import spatial.tree.lab.utility.RTreeUtility;

/**
 *
 * @author Mahmoud
 */
public class TestRTreeTex_6_2_5_tFile1000 {
    /**
     *
     * @param args
     */
    public static void main(String[] args) throws IOException, InterruptedException{
        System.out.println("********** TestRTreeTextFile1000 **************\n");

        // RTree
        //int dim = 5;
        //int m = 2;
        //int M = 5;
        
        //RTree rtree = new RTree(dim, m, M);
        //rtree.setDebugMode(0);
        
        // RTree*
        //int dim = 6;
        //int m = 2;
        //int M = 5;
        //int reInsertCount = 2;
        //RAsteriskRTree rtree = new RAsteriskRTree(dim, m, M, reInsertCount);
        //rtree.setDebugMode(0);
        
        // RTree X
        int dim = 5;
        int m = 2;
        int M = 5;
        int reInsertCount = 3;
        double maxOverlap = .2;
        int fanout =  4;
        XTree rtree = new XTree(null,dim, m, M, reInsertCount, maxOverlap, fanout);
        rtree.setDebugMode(0);
        
        // RTree +
        //int dim = 6;
        //int m = 2;
        //int M = 4;
        //int fillFactor = 3;
        //RPlusTree rtree = new RPlusTree(dim, m, M, fillFactor);
        //rtree.setDebugMode(0);
        
        SnytheticDataSetGenerator generator = new SnytheticDataSetGenerator();
        generator.createDataSet("c:\\temp\\Tree2000.txt", 5,  1000 * 1000 , 20.0, 'E');

        TextFileAdapter textFileAdapter = new TextFileAdapter();
        textFileAdapter.setFileName("c:\\temp\\Tree2000.txt");
        
        textFileAdapter.start(rtree);
        
        Node leaf = textFileAdapter.getNext(rtree);
        while(leaf != null) {
            if(leaf.getId() == "15") {
                System.out.println();
            }
            rtree.InsertData(leaf);
            leaf = textFileAdapter.getNext(rtree);
        }
        
        textFileAdapter.end();
        
        System.out.println("********** Tree **************\n");
        rtree.print();
        
        System.out.println("********** Tree Output **************\n");
        
        System.out.println("Number of Levels:" + Integer.toString(rtree.getLeafLevel()));
        //System.out.println("Nonleaf Nodes Count:" + Integer.toString(rtree.getNodeCount()));
        System.out.println("Leaves Count:" + Integer.toString(rtree.getLeafCount()));
        System.out.println("Single Node Memory Size:" + Integer.toString(rtree.getMaxNodeMemorySize()));
        System.out.println("Memory Page Size:" + Integer.toString(ByteBufferUtility.MEMORY_PAGE_SIZE));
        System.out.println("Memory Size:" + Long.toString(rtree.getMemorySize()));
        System.out.println("Hard Drive Size:" + Long.toString(rtree.getHardDriveSize()));
        System.out.println();

                
        System.out.println("********** Save the Tree **************\n");
        FileUtility fileUtility = new FileUtility();
        //fileUtility.saveToFile(rtree, "c:\\temp\\rtree_6_2_5_1000.txt");
        
        System.out.println("********** Get Root MRB **************\n");
        RTreeUtility utlity = new RTreeUtility();
        Node root = utlity.getRootNode("c:\\temp\\rtree_6_2_5_1000.txt");
        System.out.println("Root: " + root.toFullString());
        

        RandomUniformSelectivityQueryGenerator randSelectivityQuery = new RandomUniformSelectivityQueryGenerator();
        List<MinBoundingRectangle> randQueries = randSelectivityQuery.generateRandomQueries(rtree, 5, 100);
        
        
        
        System.out.println("********** Generate Random Query **************\n");        
        RandomUniformQueryGenerator randQueryGenerator = new RandomUniformQueryGenerator();
        //List<MinBoundingRectangle> randQueries = randQueryGenerator.generateRandomQueries(
        //            root.getMinBoundingRectangle(),
        //            100,
        //            2,
        //            10);

        System.out.println("********** Execute Random Query **************\n");
        for(int i = 0; i < randQueries.size(); i++) {
            System.out.println(i + "- " + randQueries.get(i).toFullString());
            
            RandomAccessFile file = fileUtility.openRandomAccessFile("c:\\temp\\rtree_6_2_5_1000.txt");
            RTree rtreeToQuery = new RTree(null,dim, m, M);
            MinBoundingRectangle rqueryMbr = randQueries.get(i);
            RQueryResult rQueryResult = rtreeToQuery.getNodes(rqueryMbr);
            rQueryResult.markNodesAsSelected();
            
            int pages = 0;
            for(Node node:rQueryResult.getVisitedNodes()) {
                pages += node.getPagesCount();
            }
            
            System.out.println("Number of Pages Visited:" + Integer.toString(pages));
            System.out.println("Number of Nodes Visited:" + Integer.toString(rQueryResult.getVisitedNodes().size()));
            System.out.println("Number of Data Leafs:" + Integer.toString(rQueryResult.getLeaves().size()));
            System.out.println("Execution Time in Nano Second:" + Long.toString(rQueryResult.getExecutionTime()));
            
            rtreeToQuery.printSelected();
            
            //for(Node data:rQueryResult.getLeaves()) {
            //    System.out.println(data.getParent().toFullString(data.getParent().getLevel()));
            //    System.out.println(data.toFullString(data.getLevel()));
            //}
            
            rQueryResult.unmarkNodesAsSelected();
            
            file.close();
        }
    }
}
