/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package spatial.tree.lab.utility;

import java.util.ArrayList;
import java.util.List;

/**
 *
 * @author Mahmoud
 */
public class FuzzyUtility {
    public double Defuzzify(double step, List<Double> range, List<Double> values, DefuzzificationMethod method) {
        double result = 0;
        
        if(method == DefuzzificationMethod.Centroid) {
            double area = 0;
            for(int i = 0; i < range.size(); i++) {
                result += range.get(i) * values.get(i) * step;
                area += values.get(i) * step;
            }
            
            if(area == 0) {
                result = (range.get(0) + range.get(range.size() - 1)) / 2;
            }
            else {
                result = result / area;
            }
        }
        
        if(method == DefuzzificationMethod.Bisector) {
            double max = Double.MAX_VALUE;
            
            for(int k = 1; k < range.size(); k++) {
                double area1 = 0;
                for(int i = 0; i < k; i ++) {
                    area1 += values.get(i) * step;
                }

                double area2 = 0;
                for(int j = k; j < range.size(); j++) {
                    area2 += values.get(j) * step;
                }
                
                if(Math.abs(area2 - area1) < max) {
                    max = Math.abs(area2 - area1);
                    
                    result = range.get(k);
                }
            }
            
            double area = 0;
            for(int i = 0; i < range.size(); i++) {
                area += values.get(i) * step;
            }
            
            if(area == 0) {
                result = (range.get(0) + range.get(range.size() - 1)) / 2;
            }
        }

        if(method == DefuzzificationMethod.LOM) {
            double max = Double.NEGATIVE_INFINITY;
            
            for(int i = 0; i < range.size(); i++) {
                if(values.get(i) >= max) {
                    max = values.get(i);
                    
                    result = range.get(i);
                }
            }
        }
        
        if(method == DefuzzificationMethod.SOM) {
            double max = Double.NEGATIVE_INFINITY;
            
            for(int i = 0; i < range.size(); i++) {
                if(values.get(i) > max) {
                    max = values.get(i);
                    
                    result = range.get(i);
                }
            }
        }

        if(method == DefuzzificationMethod.MOM) {
            double max1 = Double.NEGATIVE_INFINITY;
            double range1 = 0; 
            for(int i = 0; i < range.size(); i++) {
                if(values.get(i) >= max1) {
                    max1 = values.get(i);
                    range1 = range.get(i);
                }
            }
            
            double max2 = Double.NEGATIVE_INFINITY;
            double range2 = 0; 
            for(int i = 0; i < range.size(); i++) {
                if(values.get(i) >= max2) {
                    max2 = values.get(i);
                    range2 = range.get(i);
                }
            }
            
            result = (range1 + range2) / 2;
        }

        return result;
    }

    public static void main(String[] args) {

        
        double[] x = new double[] {1, 2, 3, 4, 5, 6, 7, 8, 9, 10 };
        List<Double> x1 = new ArrayList<>();
        for(int i = 0; i < x.length; i++) {
            x1.add(x[i]);
        }
    
        double[] v = new double[] {0, .1, .4, 1, 1, 1, .5, .5, .6, .5};
        List<Double> v1 = new ArrayList<>();
        for(int i = 0; i < v.length; i++) {
            v1.add(v[i]);
        }
        
        FuzzyUtility fu = new FuzzyUtility();
        double lom = fu.Defuzzify(1, x1, v1, DefuzzificationMethod.LOM);
        double som = fu.Defuzzify(1, x1, v1, DefuzzificationMethod.SOM);
        double mom = fu.Defuzzify(1, x1, v1, DefuzzificationMethod.MOM);
        double cen = fu.Defuzzify(1, x1, v1, DefuzzificationMethod.Centroid);
        double bi = fu.Defuzzify(1, x1, v1, DefuzzificationMethod.Bisector);
    }
}
